"""Enumeration module."""
from __future__ import annotations

from enum import IntEnum


class ParamEnum(IntEnum):
    """Wraps IntEnum to provide validation of a requested item.

    Intended for enums used for function parameters.

    Use enum.get_value(item) for this behavior instead of builtin enum[item].
    """

    @classmethod
    def get_valid_options(cls) -> dict[int, str]:
        """Get valid options as a dict."""
        return dict((e.value, e.name) for e in cls)

    @classmethod
    def get_value(cls, item: str) -> int:
        """Get integer value from str key.

        Validate item and raise a ValueError with valid options.
        """
        try:
            return cls[item].value
        except KeyError:
            valid_options_list = [
                f"'{n}' ({v})" for (v, n) in cls.get_valid_options().items()
            ]
            raise ValueError(
                "{}: '{}' is not a valid option, must be one of {}".format(
                    cls.__name__, item, ", ".join(valid_options_list)
                )
            )


class Prec(ParamEnum):
    """Floating precision type, where 1 is single and 2 is double precision."""

    single = 1
    double = 2


class KrigType(ParamEnum):
    """Kriging type, where 0 is simple and 1 is ordinary."""

    simple = 0
    ordinary = 1


class VarioType(ParamEnum):
    """Variogram type, where 1:spher, 2:exp, 3:gauss and 4:pow."""

    spher = 1
    exp = 2
    gauss = 3
    pow = 4


class FactorFileType(ParamEnum):
    """Factor file type, where 0 is binary and 1 is text."""

    binary = 0
    text = 1


class StrucType(ParamEnum):
    """Structure type, where 0 is for polylinear and 1 for polygonal."""

    polylinear = 0
    polygonal = 1


class TransType(ParamEnum):
    """Enable log-transformation of values."""

    none = 0
    log = 1



"""Mid-level pestutilslib module to implement ctypes functions."""
from __future__ import annotations

import logging
from ctypes import byref, c_char, c_double, c_int, create_string_buffer
from os import PathLike
from pathlib import Path

import numpy as np
import numpy.typing as npt

from . import enum
from .data import ManyArrays, validate_scalar


class PestUtilsLibError(BaseException):
    """Exception from PestUtilsLib."""

    pass


class PestUtilsLib:
    """Mid-level Fortran-Python handler for pestutils library via ctypes.

    Parameters
    ----------
    logger_level : int, str, default 20 (INFO)
    """

    def __init__(self, *, logger_level=logging.INFO) -> None:
        from .ctypes_declarations import prototype
        from .finder import load
        from .logger import get_logger

        self.logger = get_logger(self.__class__.__name__, logger_level)
        self.pestutils = load()
        self.logger.debug("loaded %s", self.pestutils)
        prototype(self.pestutils)
        self.logger.debug("added prototypes")

    # def __del__(self):
    #    """Clean-up library instance."""
    #    try:
    #        self.free_all_memory()
    #    except (AttributeError, PestUtilsLibError) as err:
    #        if hasattr(self, "logger"):
    #            self.logger.warning("cannot call __del__: %s", err)

    def create_char_array(self, init: str | bytes, name: str):
        """Create c_char Array with a fixed size from dimvar and intial value.

        Parameters
        ----------
        init : str or bytes
            Initial value.
        name : str
            Uppercase variable length name, e.g. LENFILENAME or LENVARTYPE.
        """
        from .ctypes_declarations import get_dimvar_int

        if isinstance(init, str):
            init = init.encode()
        elif isinstance(init, bytes):
            pass
        else:
            raise TypeError(f"expecting either str or bytes; found {type(init)}")
        size = get_dimvar_int(self.pestutils, name)
        if len(init) > size:
            raise ValueError(f"init size is {len(init)} but {name} is {size}")
        return create_string_buffer(init, size)

    def inquire_modflow_binary_file_specs(
        self,
        filein: str | PathLike,
        fileout: str | PathLike | None,
        isim: int,
        itype: int,
    ) -> dict:
        """Report some of the details of a MODFLOW-written binary file.

        Parameters
        ----------
        filein : str or PathLike
            MODFLOW-generated binary file to be read.
        fileout : str, PathLike, None
            Output file with with table of array headers. Use None or "" for
            no output file.
        isim : int
            Inform the function the simulator that generated the binary file:

             * 1 = traditional MODFLOW
             * 21 = MODFLOW-USG with structured grid
             * 22 = MODFLOW-USG with unstructured grid
             * 31 = MODFLOW 6 with DIS grid
             * 32 = MODFLOW 6 with DISV grid
             * 33 = MODFLOW 6 with DISU grid

        itype : int
            Where 1 = system state or dependent variable;
            2 = cell-by-cell flows.

        Returns
        -------
        iprec : int
            Where 1 = single; 2 = double.
        narray : int
            Number of arrays.
        ntime : int
            Number of times.
        """
        filein = Path(filein)
        if not filein.is_file():
            raise FileNotFoundError(f"could not find filein {filein}")
        if fileout:
            fileout = Path(fileout)
        else:
            fileout = b""
        validate_scalar("isim", isim, isin=[1, 21, 22, 31, 32, 33])
        validate_scalar("itype", itype, isin=[1, 2])
        iprec = c_int()
        narray = c_int()
        ntime = c_int()
        res = self.pestutils.inquire_modflow_binary_file_specs(
            byref(self.create_char_array(bytes(filein), "LENFILENAME")),
            byref(self.create_char_array(bytes(fileout), "LENFILENAME")),
            byref(c_int(isim)),
            byref(c_int(itype)),
            byref(iprec),
            byref(narray),
            byref(ntime),
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("inquired modflow binary file specs from %r", filein.name)
        return {
            "iprec": iprec.value,
            "narray": narray.value,
            "ntime": ntime.value,
        }

    def retrieve_error_message(self) -> str:
        """Retrieve error message from library.

        Returns
        -------
        str
        """
        from .ctypes_declarations import get_char_array

        charray = get_char_array(self.pestutils, "LENMESSAGE")()
        res = self.pestutils.retrieve_error_message(byref(charray))
        return charray[:res].rstrip(b"\x00").decode()

    def install_structured_grid(
        self,
        gridname: str,
        ncol: int,
        nrow: int,
        nlay: int,
        icorner: int,
        e0: float,
        n0: float,
        rotation: float,
        delr: float | npt.ArrayLike,
        delc: float | npt.ArrayLike,
    ) -> None:
        """Install specifications for a structured grid.

        Parameters
        ----------
        gridname : str
            Unique non-blank grid name.
        ncol, nrow, nlay : int
            Grid dimensions.
        icorner : int
            Reference corner, use 1 for top left and 2 for bottom left.
        e0, n0 : float
            Reference offsets.
        rotation : float
            Grid rotation, counter-clockwise degrees.
        """
        validate_scalar("ncol", ncol, gt=0)
        validate_scalar("nrow", nrow, gt=0)
        validate_scalar("nlay", nlay, gt=0)
        col = ManyArrays(float_any={"delr": delr}, ar_len=ncol)
        row = ManyArrays(float_any={"delc": delc}, ar_len=nrow)
        col.validate("delr", gt=0.0)
        row.validate("delc", gt=0.0)
        validate_scalar("icorner", icorner, isin=[1, 2])
        res = self.pestutils.install_structured_grid(
            byref(self.create_char_array(gridname, "LENGRIDNAME")),
            byref(c_int(ncol)),
            byref(c_int(nrow)),
            byref(c_int(nlay)),
            byref(c_int(icorner)),
            byref(c_double(e0)),
            byref(c_double(n0)),
            byref(c_double(rotation)),
            col.delr,
            row.delc,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("installed structured grid %r from specs", gridname)

    def get_cell_centres_structured(self, gridname: str, ncpl: int) -> tuple:
        """Get cell centres of a single layer of an installed structured grid.

        Parameters
        ----------
        gridname : str
            Name of installed structured grid.
        ncpl : int
            Dimensions of grid (nrow x ncol).

        Returns
        -------
        cellx, cellx : npt.NDArray[np.float64]
            Coordinates of cell centres with dimensions (ncpl,).
        """
        cellx = np.zeros(ncpl, np.float64, order="F")
        celly = np.zeros(ncpl, np.float64, order="F")
        res = self.pestutils.get_cell_centres_structured(
            byref(self.create_char_array(gridname, "LENGRIDNAME")),
            byref(c_int(ncpl)),
            cellx,
            celly,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info(
            "evaluated %d cell centres from structured grid %r", ncpl, gridname
        )
        return cellx.copy("A"), celly.copy("A")

    def uninstall_structured_grid(self, gridname: str) -> None:
        """Uninstall structured grid set by :meth:`install_structured_grid`.

        Parameters
        ----------
        gridname : str
            Unique non-blank grid name.
        """
        res = self.pestutils.uninstall_structured_grid(
            byref(self.create_char_array(gridname, "LENGRIDNAME"))
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("uninstalled structured grid %r", gridname)

    def free_all_memory(self) -> None:
        """Deallocate all memory that is being used."""
        ret = self.pestutils.free_all_memory()
        if ret != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("all memory was freed up")

    def interp_from_structured_grid(
        self,
        gridname: str,
        depvarfile: str | PathLike,
        isim: int,
        iprec: int | str | enum.Prec,
        ntime: int,
        vartype: str,
        interpthresh: float,
        nointerpval: float,
        # npts: int,  # determined from layer.shape[0]
        ecoord: npt.ArrayLike,
        ncoord: npt.ArrayLike,
        layer: int | npt.ArrayLike,
    ) -> dict:
        """Spatial interpolate points from a structured grid.

        Parameters
        ----------
        gridname : str
            Name of installed structured grid.
        depvarfile : str or PathLike
            Name of binary file to read.
        isim : int
            Specify -1 for MT3D; 1 for MODFLOW.
        iprec : int, str or enum.Prec
            Specify 1 or "single", 2 or "double", or use enum.Prec.
        ntime : int
            Number of output times.
        vartype : str
            Only read arrays of this type.
        interpthresh : float
            Absolute threshold for dry or inactive.
        nointerpval : float
            Value to use where interpolation is not possible.
        ecoord, ncoord : array_like
            X/Y or Easting/Northing coordinates for points with shape (npts,).
        layer : int or array_like
            Layers of points with shape (npts,).

        Returns
        -------
        nproctime : int
            Number of processed simulation times.
        simtime : npt.NDArray[np.float64]
            Simulation times, with shape (ntime,).
        simstate : npt.NDArray[np.float64]
            Interpolated system states, with shape (ntime, npts).
        """
        depvarfile = Path(depvarfile)
        if not depvarfile.is_file():
            raise FileNotFoundError(f"could not find depvarfile {depvarfile}")
        if isinstance(iprec, str):
            iprec = enum.Prec.get_value(iprec)
        pta = ManyArrays({"ecoord": ecoord, "ncoord": ncoord}, int_any={"layer": layer})
        npts = len(pta)
        simtime = np.zeros(ntime, np.float64, order="F")
        simstate = np.zeros((ntime, npts), np.float64, order="F")
        nproctime = c_int()
        res = self.pestutils.interp_from_structured_grid(
            byref(self.create_char_array(gridname, "LENGRIDNAME")),
            byref(self.create_char_array(bytes(depvarfile), "LENFILENAME")),
            byref(c_int(isim)),
            byref(c_int(iprec)),
            byref(c_int(ntime)),
            byref(self.create_char_array(vartype, "LENVARTYPE")),
            byref(c_double(interpthresh)),
            byref(c_double(nointerpval)),
            byref(c_int(npts)),
            pta.ecoord,
            pta.ncoord,
            pta.layer,
            byref(nproctime),
            simtime,
            simstate,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info(
            "interpolated %d points from structured grid %r", npts, gridname
        )
        return {
            "nproctime": nproctime.value,
            "simtime": simtime.copy("A"),
            "simstate": simstate.copy("A"),
        }

    def interp_to_obstime(
        self,
        # nsimtime: int,  # determined from simval.shape[0]
        nproctime: int,
        # npts: int,  # determined from simval.shape[1]
        simtime: npt.ArrayLike,
        simval: npt.ArrayLike,
        interpthresh: float,
        how_extrap: str,
        time_extrap: float,
        nointerpval: float,
        # nobs: int,  # determined from obspoint.shape[0]
        obspoint: npt.ArrayLike,
        obstime: npt.ArrayLike,
    ) -> npt.NDArray[np.float64]:
        """Temporal interpolation for simulation times to observed times.

        Parameters
        ----------
        nproctime : int
            Number of times featured in simtime and simval.
        simtime : array_like
            1D array of simulation times with shape (nsimtime,).
        simval : array_like
            2D array of simulated values with shape (nsimtime, npts).
        interpthresh : float
            Values equal or above this in simval have no meaning.
        how_extrap : str
            Method, where 'L'=linear; 'C'=constant.
        time_extrap : float
            Permitted extrapolation time.
        nointerpval : float
            Value to use where interpolation is not possible.
        obspoint : array_like
            1D integer array of indices of observation points,
            which start at 0 and -1 means no index. Shape is (nobs,).
        obstime : array_like
            1D array of observation times with shape (nobs,).

        Returns
        -------
        np.ndarray
            Time-interpolated simulation values with shape (nobs,).
        """
        simtime = np.array(simtime, dtype=np.float64, order="F", copy=False)
        simval = np.array(simval, dtype=np.float64, order="F", copy=False)
        obspoint = np.array(obspoint, order="F", copy=False)
        obstime = np.array(obstime, dtype=np.float64, order="F", copy=False)
        if simtime.ndim != 1:
            raise ValueError("expected 'simtime' to have ndim=1")
        elif simval.ndim != 2:
            raise ValueError("expected 'simval' to have ndim=2")
        elif obspoint.ndim != 1:
            raise ValueError("expected 'obspoint' to have ndim=1")
        elif obstime.ndim != 1:
            raise ValueError("expected 'obstime' to have ndim=1")
        elif not np.issubdtype(obspoint.dtype, np.integer):
            raise ValueError(
                f"expected 'obspoint' to be integer type; found {obspoint.dtype}"
            )
        nsimtime, npts = simval.shape
        nobs = len(obspoint)
        obssimval = np.zeros(nobs, np.float64, order="F")
        res = self.pestutils.interp_to_obstime(
            byref(c_int(nsimtime)),
            byref(c_int(nproctime)),
            byref(c_int(npts)),
            simtime,
            simval,
            byref(c_double(interpthresh)),
            byref(c_char(how_extrap.encode())),
            byref(c_double(time_extrap)),
            byref(c_double(nointerpval)),
            byref(c_int(nobs)),
            obspoint.astype(np.int32, copy=False),
            obstime,
            obssimval,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("interpolated %d time points to %d observations", npts, nobs)
        return obssimval.copy("A")

    def install_mf6_grid_from_file(
        self, gridname: str, grbfile: str | PathLike
    ) -> dict:
        """Install specifications for a MF6 grid from a GRB file.

        Parameters
        ----------
        gridname : str
            Unique non-blank grid name.
        grbfile : str or PathLike
            Path to a GRB binary grid file.

        Returns
        -------
        idis : int
            Where 1 is for DIS and 2 is for DISV.
        ncells : int
            Number of cells in the grid.
        ndim1, ndim2, ndim3 : int
            Grid dimensions.
        """
        grbfile = Path(grbfile)
        if not grbfile.is_file():
            raise FileNotFoundError(f"could not find grbfile {grbfile}")
        idis = c_int()
        ncells = c_int()
        ndim1 = c_int()
        ndim2 = c_int()
        ndim3 = c_int()
        res = self.pestutils.install_mf6_grid_from_file(
            byref(self.create_char_array(gridname, "LENGRIDNAME")),
            byref(self.create_char_array(bytes(grbfile), "LENFILENAME")),
            byref(idis),
            byref(ncells),
            byref(ndim1),
            byref(ndim2),
            byref(ndim3),
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info(
            "installed mf6 grid %r from grbfile=%r", gridname, grbfile.name
        )
        return {
            "idis": idis.value,
            "ncells": ncells.value,
            "ndim1": ndim1.value,
            "ndim2": ndim2.value,
            "ndim3": ndim3.value,
        }

    def get_cell_centres_mf6(self, gridname: str, ncells: int) -> tuple:
        """Get cell centres from an installed MF6 grid.

        Parameters
        ----------
        gridname : str
            Name of installed MF6 grid.
        ncells : int
            Dimensions of grid.

        Returns
        -------
        cellx, cellx, cellz : npt.NDArray[np.float64]
            Coordinates of cell centres with dimensions (ncells,).
        """
        cellx = np.zeros(ncells, np.float64, order="F")
        celly = np.zeros(ncells, np.float64, order="F")
        cellz = np.zeros(ncells, np.float64, order="F")
        res = self.pestutils.get_cell_centres_mf6(
            byref(self.create_char_array(gridname, "LENGRIDNAME")),
            byref(c_int(ncells)),
            cellx,
            celly,
            cellz,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("evaluated %d cell centres from MF6 grid %r", ncells, gridname)
        return cellx.copy("A"), celly.copy("A"), cellz.copy("A")

    def uninstall_mf6_grid(self, gridname: str) -> None:
        """Uninstall MF6 grid set by :meth:`install_mf6_grid_from_file`.

        Parameters
        ----------
        gridname : str
            Unique non-blank grid name.
        """
        res = self.pestutils.uninstall_mf6_grid(
            byref(self.create_char_array(gridname, "LENGRIDNAME"))
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("uninstalled mf6 grid %r", gridname)

    def calc_mf6_interp_factors(
        self,
        gridname: str,
        # npts: int,  # determined from ecoord.shape[0]
        ecoord: npt.ArrayLike,
        ncoord: npt.ArrayLike,
        layer: int | npt.ArrayLike,
        factorfile: str | PathLike,
        factorfiletype: int | str | enum.FactorFileType,
        blnfile: str | PathLike,
    ) -> npt.NDArray[np.int32]:
        """Calculate interpolation factors from a MODFLOW 6 DIS or DISV.

        Parameters
        ----------
        gridname : str
            Unique non-blank grid name.
        ecoord, ncoord : array_like
            X/Y or Easting/Northing coordinates for points with shape (npts,).
        layer : int or array_like
            Layers of points with shape (npts,).
        factorfile : str or PathLike
            File for kriging factors to write.
        factorfiletype : int, str or enum.FactorFileType
            Factor file type, where 0:binary, 1:text.
        blnfile : str or PathLike
            Name of bln file to write.

        Returns
        -------
        npt.NDArray[np.int32]
            Array interp_success(npts), where 1 is success and 0 is failure.
        """
        pta = ManyArrays({"ecoord": ecoord, "ncoord": ncoord}, int_any={"layer": layer})
        npts = len(pta)
        factorfile = Path(factorfile)
        if isinstance(factorfiletype, str):
            factorfiletype = enum.FactorFileType.get_value(factorfiletype)
        blnfile = Path(blnfile)
        interp_success = np.zeros(npts, np.int32, order="F")
        res = self.pestutils.calc_mf6_interp_factors(
            byref(self.create_char_array(gridname, "LENGRIDNAME")),
            byref(c_int(npts)),
            pta.ecoord,
            pta.ncoord,
            pta.layer,
            byref(self.create_char_array(bytes(factorfile), "LENFILENAME")),
            byref(c_int(factorfiletype)),
            byref(self.create_char_array(bytes(blnfile), "LENFILENAME")),
            interp_success,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("calculated mf6 interp factors for %r", gridname)
        return interp_success.copy("A")

    def interp_from_mf6_depvar_file(
        self,
        depvarfile: str | PathLike,
        factorfile: str | PathLike,
        factorfiletype: int | str | enum.FactorFileType,
        ntime: int,
        vartype: str,
        interpthresh: float,
        reapportion: int | bool,
        nointerpval: float,
        npts: int,
    ) -> dict:
        """
        Interpolate points using previously-calculated interpolation factors.

        Parameters
        ----------
        depvarfile : str or PathLike
            Name of binary file to read.
        factorfile : str or PathLike
            File containing spatial interpolation factors, written by
            :meth:`calc_mf6_interp_factors`.
        factorfiletype : int, str or enum.FactorFileType
            Use 0 for binary; 1 for text.
        ntime : int
            Number of output times.
        vartype : str
            Only read arrays of this type.
        interpthresh : float
            Absolute threshold for dry or inactive.
        reapportion : int or bool
            Use 0 for no (False); 1 for yes (True).
        nointerpval : float
            Value to use where interpolation is not possible.
        npts : int
            Number of points for interpolation.

        Returns
        -------
        nproctime : int
            Number of processed simulation times.
        simtime : npt.NDArray[np.float64]
            Simulation times, with shape (ntime,).
        simstate : npt.NDArray[np.float64]
            Interpolated system states, with shape (ntime, npts).
        """
        depvarfile = Path(depvarfile)
        if not depvarfile.is_file():
            raise FileNotFoundError(f"could not find depvarfile {depvarfile}")
        factorfile = Path(factorfile)
        if not factorfile.is_file():
            raise FileNotFoundError(f"could not find factorfile {factorfile}")
        if isinstance(factorfiletype, str):
            factorfiletype = enum.FactorFileType.get_value(factorfiletype)
        simtime = np.zeros(ntime, np.float64, order="F")
        simstate = np.zeros((ntime, npts), np.float64, order="F")
        nproctime = c_int()
        res = self.pestutils.interp_from_mf6_depvar_file(
            byref(self.create_char_array(bytes(depvarfile), "LENFILENAME")),
            byref(self.create_char_array(bytes(factorfile), "LENFILENAME")),
            byref(c_int(factorfiletype)),
            byref(c_int(ntime)),
            byref(self.create_char_array(vartype, "LENVARTYPE")),
            byref(c_double(interpthresh)),
            byref(c_int(reapportion)),
            byref(c_double(nointerpval)),
            byref(c_int(npts)),
            byref(nproctime),
            simtime,
            simstate,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info(
            "interpolated %d points from mf6 depvar file %r", npts, depvarfile.name
        )
        return {
            "nproctime": nproctime.value,
            "simtime": simtime.copy("A"),
            "simstate": simstate.copy("A"),
        }

    def extract_flows_from_cbc_file(
        self,
        cbcfile: str | PathLike,
        flowtype: str,
        isim: int,
        iprec: int | str | enum.Prec,
        # ncell: int,  # from izone.shape[0]
        izone: npt.ArrayLike,
        nzone: int,
        ntime: int,
    ) -> dict:
        """
        Read and accumulates flows from a CBC flow file to a user-specified BC.

        Parameters
        ----------
        cbcfile : str | PathLike
            Cell-by-cell flow term file written by any MF version.
        flowtype : str
            Type of flow to read.
        isim : int
            Simulator type.
        iprec : int, str or enum.Prec
            Precision used to record real variables in cbc file.
        izone : array_like
            Zonation of model domain, with shape (ncell,).
        nzone : int
            Equals or exceeds number of zones; zone 0 doesn't count.
        ntime : int
            Equals or exceed number of model output times for flow type.

        Returns
        -------
        numzone : int
            Number of non-zero-valued zones.
        zonenumber : npt.NDArray[np.int32]
            Zone numbers, with shape (nzone,).
        nproctime : int
            Number of processed simulation times.
        timestep : npt.NDArray[np.int32]
            Simulation time step, with shape (ntime,).
        stressperiod : npt.NDArray[np.int32]
            Simulation stress period, with shape (ntime,).
        simtime : npt.NDArray[np.int32]
            Simulation time, with shape (ntime,).
            A time of -1.0 indicates unknown.
        simflow : npt.NDArray[np.int32]
            Interpolated flows, with shape (ntime, nzone).
        """
        cbcfile = Path(cbcfile)
        if not cbcfile.is_file():
            raise FileNotFoundError(f"could not find cbcfile {cbcfile}")
        validate_scalar("flowtype", flowtype, minlen=1)
        validate_scalar("iprec", iprec, enum=enum.Prec)
        if isinstance(iprec, str):
            iprec = enum.Prec.get_value(iprec)
        cell = ManyArrays(int_any={"izone": izone})
        ncell = len(cell)
        numzone = c_int()
        zonenumber = np.zeros(nzone, np.int32, order="F")
        nproctime = c_int()
        timestep = np.zeros(ntime, np.int32, order="F")
        stressperiod = np.zeros(ntime, np.int32, order="F")
        simtime = np.zeros(ntime, np.float64, order="F")
        simflow = np.zeros((ntime, nzone), np.float64, order="F")
        res = self.pestutils.extract_flows_from_cbc_file(
            byref(self.create_char_array(bytes(cbcfile), "LENFILENAME")),
            byref(self.create_char_array(flowtype, "LENFLOWTYPE")),
            byref(c_int(isim)),
            byref(c_int(iprec)),
            byref(c_int(ncell)),
            cell.izone,
            byref(c_int(nzone)),
            byref(numzone),
            zonenumber,
            byref(c_int(ntime)),
            byref(nproctime),
            timestep,
            stressperiod,
            simtime,
            simflow,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("extracted flows from %r", cbcfile.name)
        return {
            "numzone": numzone.value,
            "zonenumber": zonenumber.copy("A"),
            "nproctime": nproctime.value,
            "timestep": timestep.copy("A"),
            "stressperiod": stressperiod.copy("A"),
            "simtime": simtime.copy("A"),
            "simflow": simflow.copy("A"),
        }

    def calc_kriging_factors_2d(
        self,
        # npts: int,  # determined from ecs.shape[0]
        ecs: npt.ArrayLike,
        ncs: npt.ArrayLike,
        zns: int | npt.ArrayLike,
        # mpts: int,  # determined from ect.shape[0]
        ect: npt.ArrayLike,
        nct: npt.ArrayLike,
        znt: int | npt.ArrayLike,
        vartype: int | str | enum.VarioType,
        krigtype: int | str | enum.KrigType,
        aa: float | npt.ArrayLike,
        anis: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        searchrad: float,
        maxpts: int,
        minpts: int,
        factorfile: str | PathLike,
        factorfiletype: int | str | enum.FactorFileType,
    ) -> int:
        """
        Calculate 2D kriging factors.

        Parameters
        ----------
        ecs, ncs : array_like
            Source point coordinates, each 1D array with shape (npts,).
        zns : int or array_like
            Source point zones, integer or 1D array with shape (npts,).
        ect, nct : array_like
            Target point coordinates, each 1D array with shape (mpts,).
        znt : int or array_like
            Target point zones, integer or 1D array with shape (mpts,).
        vartype : int, str or enum.VarioType
            Variogram type, where 1:spher, 2:exp, 3:gauss, 4:pow.
        krigtype : int, str, or enum.KrigType,
            Kriging type, where 0:simple, 1:ordinary.
        aa : float or array_like
            Variogram "a" value, float or 1D array with shape (mpts,).
        anis : float or array_like
            Variogram anisotropies, float or 1D array with shape (mpts,).
        bearing : float or array_like
            Variogram bearings, float or 1D array with shape (mpts,).
        searchrad : float
            Search radius.
        maxpts, minpts : int
            Search specifications.
        factorfile : str or PathLike
            File for kriging factors.
        factorfiletype : int, str or enum.FactorFileType
            Factor file type, where 0:binary, 1:text.

        Returns
        -------
        int
            Number of interp points.
        """
        npta = ManyArrays({"ecs": ecs, "ncs": ncs}, int_any={"zns": zns})
        npts = len(npta)
        mpta = ManyArrays(
            {"ect": ect, "nct": nct},
            {"aa": aa, "anis": anis, "bearing": bearing},
            {"znt": znt},
        )
        mpts = len(mpta)
        if isinstance(vartype, str):
            vartype = enum.VarioType.get_value(vartype)
        if isinstance(krigtype, str):
            krigtype = enum.KrigType.get_value(krigtype)
        factorfile = Path(factorfile)
        if isinstance(factorfiletype, str):
            factorfiletype = enum.FactorFileType.get_value(factorfiletype)
        icount_interp = c_int()
        res = self.pestutils.calc_kriging_factors_2d(
            byref(c_int(npts)),
            npta.ecs,
            npta.ncs,
            npta.zns,
            byref(c_int(mpts)),
            mpta.ect,
            mpta.nct,
            mpta.znt,
            byref(c_int(vartype)),
            byref(c_int(krigtype)),
            mpta.aa,
            mpta.anis,
            mpta.bearing,
            byref(c_double(searchrad)),
            byref(c_int(maxpts)),
            byref(c_int(minpts)),
            byref(self.create_char_array(bytes(factorfile), "LENFILENAME")),
            byref(c_int(factorfiletype)),
            byref(icount_interp),
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("calculated 2D kriging factors to %r", factorfile.name)
        return icount_interp.value

    def calc_kriging_factors_auto_2d(
        self,
        # npts: int,  # determined from ecs.shape[0]
        ecs: npt.ArrayLike,
        ncs: npt.ArrayLike,
        zns: int | npt.ArrayLike,
        # mpts: int,  # determined from ect.shape[0]
        ect: npt.ArrayLike,
        nct: npt.ArrayLike,
        znt: int | npt.ArrayLike,
        krigtype: int | str | enum.KrigType,
        anis: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        factorfile: str | PathLike,
        factorfiletype: int | str | enum.FactorFileType,
    ) -> int:
        """
        Calculate 2D kriging factors, with automatic variogram properties.

        Parameters
        ----------
        ecs, ncs : array_like
            Source point coordinates, each 1D array with shape (npts,).
        zns : int or array_like
            Source point zones, integer or 1D array with shape (npts,).
        ect, nct : array_like
            Target point coordinates, each 1D array with shape (mpts,).
        znt : int or array_like
            Target point zones, integer or 1D array with shape (mpts,).
        krigtype : int, str, enum.KrigType
            Kriging type, where 0:simple, 1:ordinary.
        anis : float or array_like
            Variogram anisotropies, float or 1D array with shape (mpts,).
        bearing : float or array_like
            Variogram bearings, float or 1D array with shape (mpts,).
        factorfile : str or PathLike
            File for kriging factors.
        factorfiletype : int, str or enum.FactorFileType
            Factor file type, where 0:binary, 1:text.

        Returns
        -------
        int
            Number of interp points.
        """
        npta = ManyArrays({"ecs": ecs, "ncs": ncs}, int_any={"zns": zns})
        npts = len(npta)
        mpta = ManyArrays(
            {"ect": ect, "nct": nct}, {"anis": anis, "bearing": bearing}, {"znt": znt}
        )
        mpts = len(mpta)
        if isinstance(krigtype, str):
            krigtype = enum.KrigType.get_value(krigtype)
        factorfile = Path(factorfile)
        if isinstance(factorfiletype, str):
            factorfiletype = enum.FactorFileType.get_value(factorfiletype)
        icount_interp = c_int()
        res = self.pestutils.calc_kriging_factors_auto_2d(
            byref(c_int(npts)),
            npta.ecs,
            npta.ncs,
            npta.zns,
            byref(c_int(mpts)),
            mpta.ect,
            mpta.nct,
            mpta.znt,
            byref(c_int(krigtype)),
            mpta.anis,
            mpta.bearing,
            byref(self.create_char_array(bytes(factorfile), "LENFILENAME")),
            byref(c_int(factorfiletype)),
            byref(icount_interp),
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("calculated 2D auto kriging factors to %r", factorfile.name)
        return icount_interp.value

    def calc_kriging_factors_3d(
        self,
        # npts: int,  # determined from ecs.shape[0]
        ecs: npt.ArrayLike,
        ncs: npt.ArrayLike,
        zcs: npt.ArrayLike,
        zns: int | npt.ArrayLike,
        # mpts: int,  # determined from ect.shape[0]
        ect: npt.ArrayLike,
        nct: npt.ArrayLike,
        zct: npt.ArrayLike,
        znt: int | npt.ArrayLike,
        zonenum: int | npt.ArrayLike,
        krigtype: int | str | enum.KrigType,
        # nzone: int,  # determined from shape[0] from any zonenum..rake else 1
        vartype: int | str | enum.VarioType | npt.ArrayLike,
        ahmax: float | npt.ArrayLike,
        ahmin: float | npt.ArrayLike,
        avert: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        dip: float | npt.ArrayLike,
        rake: float | npt.ArrayLike,
        srhmax: float,
        srhmin: float,
        srvert: float,
        maxpts: int,
        minpts: int,
        factorfile: str | PathLike,
        factorfiletype: int | str | enum.FactorFileType,
    ) -> int:
        """
        Calculate 3D kriging factors.

        Parameters
        ----------
        ecs, ncs, zcs : array_like
            Source point coordinates, each 1D array with shape (npts,).
        zns : int or array_like
            Source point zones, integer or 1D array with shape (npts,).
        ect, nct, zct : array_like
            Target point coordinates, each 1D array with shape (mpts,).
        znt : int or array_like
            Target point zones, integer or 1D array with shape (mpts,).
        krigtype : int, str, or enum.KrigType,
            Kriging type, where 0:simple, 1:ordinary.
        zonenum : int, or array_like
            Zone numbers, inteter or 1D array with shape (nzone,).
        vartype : int, str, enum.VarioType or array_like
            Variogram type, where 1:spher, 2:exp, 3:gauss, 4:pow. If array,
            then it should have shape (nzone,).
        ahmax, ahmin, avert : float or array_like
            Variogram "a" values in 3 orthogonal directions (hmax, hmin, vert).
            Each can be a float or 1D array with shape (nzone,).
        bearing : float or array_like
            Bearing of hmax, float or 1D array with shape (nzone,).
        dip : float or array_like
            Dip of hmax, float or 1D array with shape (nzone,).
        rake : float or array_like
            Twist about hmax axis, float or 1D array with shape (nzone,).
        srhmax, srhmin, srvert : float
            Search radius in hmax, hmin, and vert directions.
        maxpts, minpts : int
            Search specifications.
        factorfile : str or PathLike
            File for kriging factors.
        factorfiletype : int, str or enum.FactorFileType
            Factor file type, where 0:binary, 1:text.

        Returns
        -------
        int
            Number of interp points.
        """
        npta = ManyArrays({"ecs": ecs, "ncs": ncs, "zcs": zcs}, int_any={"zns": zns})
        npts = len(npta)
        mpta = ManyArrays({"ect": ect, "nct": nct, "zct": zct}, int_any={"znt": znt})
        mpts = len(mpta)
        if isinstance(krigtype, str):
            krigtype = enum.KrigType.get_value(krigtype)
        vartype = np.array(vartype)
        if np.issubdtype(vartype.dtype, str):
            vartype = np.vectorize(enum.VarioType.get_value)(vartype)
        if not np.issubdtype(vartype.dtype, np.integer):
            raise ValueError("expected 'vartype' to be integer, str or enum.VarioType")
        nzone = ManyArrays(
            float_any={
                "ahmax": ahmax,
                "ahmin": ahmin,
                "avert": avert,
                "bearing": bearing,
                "dip": dip,
                "rake": rake,
            },
            int_any={"zonenum": zonenum, "vartype": vartype},
        )
        factorfile = Path(factorfile)
        if isinstance(factorfiletype, str):
            factorfiletype = enum.FactorFileType.get_value(factorfiletype)
        icount_interp = c_int()
        res = self.pestutils.calc_kriging_factors_3d(
            byref(c_int(npts)),
            npta.ecs,
            npta.ncs,
            npta.zcs,
            npta.zns,
            byref(c_int(mpts)),
            mpta.ect,
            mpta.nct,
            mpta.zct,
            mpta.znt,
            byref(c_int(krigtype)),
            byref(c_int(len(nzone))),
            nzone.zonenum,
            nzone.vartype,
            nzone.ahmax,
            nzone.ahmin,
            nzone.avert,
            nzone.bearing,
            nzone.dip,
            nzone.rake,
            byref(c_double(srhmax)),
            byref(c_double(srhmin)),
            byref(c_double(srvert)),
            byref(c_int(maxpts)),
            byref(c_int(minpts)),
            byref(self.create_char_array(bytes(factorfile), "LENFILENAME")),
            byref(c_int(factorfiletype)),
            byref(icount_interp),
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("calculated 3D kriging factors to %r", factorfile.name)
        return icount_interp.value

    def krige_using_file(
        self,
        factorfile: str | PathLike,
        factorfiletype: int | str | enum.FactorFileType,
        # npts: int,  # determined from sourceval.shape[0]
        mpts: int,
        krigtype: int | str | enum.KrigType,
        transtype: int | str | enum.TransType,
        sourceval: npt.ArrayLike,
        meanval: float | npt.ArrayLike | None,
        nointerpval: float,
    ) -> dict:
        """
        Apply interpolation factors calculated by other functions.

        Parameters
        ----------
        factorfile : str or PathLike
            Input file with kriging factors.
        factorfiletype : int, str or enum.FactorFileType
            Factor file type, where 0:binary, 1:text.
        mpts : int
            Number of target points, used to compare with value in factor file.
        krigtype : int, str, or enum.KrigType,
            Kriging type, where 0:simple, 1:ordinary.
        transtype : int, str, enum.TransType
            Tranformation type, where 0 is none and 1 is log.
        sourceval : array_like
            Values at sources, 1D array with shape (npts,).
        meanval : float, array_like, optional
            Mean values are required if simple kriging, described as a float
            or 1D array with shape (mpts,).
        nointerpval : float
            Value to use where interpolation is not possible.

        Returns
        -------
        targval : npt.NDArray[np.float64]
            Values calculated for targets.
        icount_interp : int
            Number of interpolation pts.
        """
        factorfile = Path(factorfile)
        if not factorfile.is_file():
            raise FileNotFoundError(f"could not find factorfile {factorfile}")
        if isinstance(factorfiletype, str):
            factorfiletype = enum.FactorFileType.get_value(factorfiletype)
        if isinstance(krigtype, str):
            krigtype = enum.KrigType.get_value(krigtype)
        if isinstance(transtype, str):
            transtype = enum.TransType.get_value(transtype)
        npta = ManyArrays({"sourceval": sourceval})
        npts = len(npta)
        if meanval is None:
            if krigtype == enum.KrigType.simple:
                self.logger.error(
                    "simple kriging requires 'meanval'; assuming zero for now"
                )
            meanval = 0.0
        mpta = ManyArrays(float_any={"meanval": meanval}, ar_len=mpts)
        targval = np.full(mpts, nointerpval, dtype=np.float64, order="F")
        icount_interp = c_int()
        res = self.pestutils.krige_using_file(
            byref(self.create_char_array(bytes(factorfile), "LENFILENAME")),
            byref(c_int(factorfiletype)),
            byref(c_int(npts)),
            byref(c_int(mpts)),
            byref(c_int(krigtype)),
            byref(c_int(transtype)),
            npta.sourceval,
            targval,
            byref(icount_interp),
            mpta.meanval,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("kriged using factor file %r", factorfile.name)
        return {
            "targval": targval.copy("A"),
            "icount_interp": icount_interp.value,
        }

    def build_covar_matrix_2d(
        self,
        # npts: int,  # determined from ec.shape[0]
        ec: npt.ArrayLike,
        nc: npt.ArrayLike,
        zn: int | npt.ArrayLike,
        vartype: int | str | enum.VarioType,
        nugget: float | npt.ArrayLike,
        aa: float | npt.ArrayLike,
        sill: float | npt.ArrayLike,
        anis: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        ldcovmat: int,
    ) -> npt.NDArray[np.float64]:
        """
        Calculate a covariance matrix for a set of 2D pilot points.

        Parameters
        ----------
        ec, nc : array_like
            Pilot point coordinates, each 1D array with shape (npts,).
        zn : int or array_like
            Pilot point zones, integer or 1D array with shape (npts,).
        vartype : int, str or enum.VarioType
            Variogram type, where 1:spher, 2:exp, 3:gauss, 4:pow.
        nugget, aa, sill, anis, bearing : float or array_like
            Variogram parameters, each float or 1D array with shape (npts,).
        ldcovmat : int
            Leading dimension of covmat.

        Returns
        -------
        npt.NDArray[np.float64]
            2D matrix covmat(ldcovmat, npts).
        """
        pta = ManyArrays(
            {"ec": ec, "nc": nc},
            {
                "nugget": nugget,
                "aa": aa,
                "sill": sill,
                "anis": anis,
                "bearing": bearing,
            },
            {"zn": zn},
        )
        npts = len(pta)
        if isinstance(vartype, str):
            vartype = enum.VarioType.get_value(vartype)
        covmat = np.zeros((ldcovmat, npts), np.float64, order="F")
        res = self.pestutils.build_covar_matrix_2d(
            byref(c_int(npts)),
            pta.ec,
            pta.nc,
            pta.zn,
            byref(c_int(vartype)),
            pta.nugget,
            pta.aa,
            pta.sill,
            pta.anis,
            pta.bearing,
            byref(c_int(ldcovmat)),
            covmat,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("calculated covariance matrix for %d 2D pilot points", npts)
        return covmat.copy("A")

    def build_covar_matrix_3d(
        self,
        # npts: int,  # determined from ec.shape[0]
        ec: npt.ArrayLike,
        nc: npt.ArrayLike,
        zc: npt.ArrayLike,
        zn: int | npt.ArrayLike,
        vartype: int | str | enum.VarioType,
        nugget: float | npt.ArrayLike,
        sill: float | npt.ArrayLike,
        ahmax: float | npt.ArrayLike,
        ahmin: float | npt.ArrayLike,
        avert: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        dip: float | npt.ArrayLike,
        rake: float | npt.ArrayLike,
        ldcovmat: int,
    ) -> npt.NDArray[np.float64]:
        """
        Calculate a covariance matrix for a set of 3D pilot points.

        Parameters
        ----------
        ec, nc, zc: array_like
            Pilot point coordinates, each 1D array with shape (npts,).
        zn : int or array_like
            Pilot point zones, integer or 1D array with shape (npts,).
        vartype : int, str or enum.VarioType
            Variogram type, where 1:spher, 2:exp, 3:gauss, 4:pow.
        nugget, sill : float or array_like
            Variogram parameters, each float or 1D array with shape (npts,).
        ahmax, ahmin, avert : float or array_like
            Variogram a parameters, each float or 1D array with shape (npts,).
        bearing, dip, rake : float or array_like
            Variogram angles, each float or 1D array with shape (npts,).
        ldcovmat : int
            Leading dimension of covmat.

        Returns
        -------
        npt.NDArray[np.float64]
            2D matrix covmat(ldcovmat, npts).
        """
        pta = ManyArrays(
            {"ec": ec, "nc": nc, "zc": zc},
            {
                "nugget": nugget,
                "sill": sill,
                "ahmax": ahmax,
                "ahmin": ahmin,
                "avert": avert,
                "bearing": bearing,
                "dip": dip,
                "rake": rake,
            },
            {"zn": zn},
        )
        npts = len(pta)
        if isinstance(vartype, str):
            vartype = enum.VarioType.get_value(vartype)
        covmat = np.zeros((ldcovmat, npts), np.float64, order="F")
        res = self.pestutils.build_covar_matrix_3d(
            byref(c_int(npts)),
            pta.ec,
            pta.nc,
            pta.zc,
            pta.zn,
            byref(c_int(vartype)),
            pta.nugget,
            pta.sill,
            pta.ahmax,
            pta.ahmin,
            pta.avert,
            pta.bearing,
            pta.dip,
            pta.rake,
            byref(c_int(ldcovmat)),
            covmat,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("calculated covariance matrix for %d 3D pilot points", npts)
        return covmat.copy("A")

    def calc_structural_overlay_factors(
        self,
        # npts: int,  # determined from ecs.shape[0]
        ecs: npt.ArrayLike,
        ncs: npt.ArrayLike,
        ids: int | npt.ArrayLike,
        conwidth: npt.ArrayLike,
        aa: npt.ArrayLike,
        structype: int | str | enum.StrucType,
        inverse_power: float,
        # mpts: int,  # determined from ect.shape[0]
        ect: npt.ArrayLike,
        nct: npt.ArrayLike,
        active: int | npt.ArrayLike,
        factorfile: str | PathLike,
        factorfiletype: int | str | enum.FactorFileType,
    ) -> int:
        """
        Calculate interpolation/blending factors for structural overlay parameters.

        Parameters
        ----------
        ecs, ncs : array_like
            Source point coordinates, each 1D array with shape (npts,).
        ids : int or array_like
            Source point structure number, integer or 1D array with shape (npts,).
        conwidth, aa : float or array_like
            Blending parameters, float or 1D array with shape (npts,).
        structype : int, str or enum.StrucType
            Structure type, where 0 is polylinear and 1 is polygonal.
        inverse_power : float
            Inverse power of distance.
        ect, nct : array_like
            Target point coordinates, each 1D array with shape (mpts,).
        active : int or array_like
            Target point activity, integer or 1D array with shape (mpts,).
        factorfile : str or PathLike
            File for kriging factors.
        factorfiletype : int, str or enum.FactorFileType
            Factor file type, where 0:binary, 1:text.

        Returns
        -------
        int
            Number of interp points.
        """
        npta = ManyArrays(
            {"ecs": ecs, "ncs": ncs}, {"conwidth": conwidth, "aa": aa}, {"ids": ids}
        )
        npts = len(npta)
        mpta = ManyArrays({"ect": ect, "nct": nct}, int_any={"active": active})
        mpts = len(mpta)
        if isinstance(structype, str):
            structype = enum.StrucType.get_value(structype)
        factorfile = Path(factorfile)
        if isinstance(factorfiletype, str):
            factorfiletype = enum.FactorFileType.get_value(factorfiletype)
        icount_interp = c_int()
        res = self.pestutils.calc_structural_overlay_factors(
            byref(c_int(npts)),
            npta.ecs,
            npta.ncs,
            npta.ids,
            npta.conwidth,
            npta.aa,
            byref(c_int(structype)),
            byref(c_double(inverse_power)),
            byref(c_int(mpts)),
            mpta.ect,
            mpta.nct,
            mpta.active,
            byref(self.create_char_array(bytes(factorfile), "LENFILENAME")),
            byref(c_int(factorfiletype)),
            byref(icount_interp),
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info(
            "calculated interpolation/blending factors to %r", factorfile.name
        )
        return icount_interp.value

    def interpolate_blend_using_file(
        self,
        factorfile: str | PathLike,
        factorfiletype: int | str | enum.FactorFileType,
        # npts: int,  # determined from sourceval.shape[0]
        # mpts: int,  # determined from targval.shape[0]
        transtype: int | str | enum.TransType,
        lt_target: str | bool,
        gt_target: str | bool,
        sourceval: npt.ArrayLike,
        targval: npt.ArrayLike,
    ) -> dict:
        """
        Apply interpolation factors calculated by :meth:`calc_structural_overlay_factors`.

        Parameters
        ----------
        factorfile : str or PathLike
            File for kriging factors.
        factorfiletype : int, str or enum.FactorFileType
            Factor file type, where 0:binary, 1:text.
        transtype : int, str, enum.TransType
            Tranformation type, where 0 is none and 1 is log.
        lt_target, gt_target : str or bool
            Whether to undercut or exceed target, use "Y"/"N" or bool.
        sourceval : array_like
            Values at sources, 1D array with shape (npts,).
        targval : array_like
            Values at targets, 1D array with shape (mpts,).

        Returns
        -------
        targval : npt.NDArray[np.float64]
            Values calculated for targets.
        icount_interp : int
            Number of interpolation pts.
        """
        factorfile = Path(factorfile)
        if not factorfile.is_file():
            raise FileNotFoundError(f"could not find factorfile {factorfile}")
        if isinstance(factorfiletype, str):
            factorfiletype = enum.FactorFileType.get_value(factorfiletype)
        if isinstance(transtype, str):
            transtype = enum.TransType.get_value(transtype)
        if isinstance(lt_target, bool):
            lt_target = "y" if lt_target else "n"
        if isinstance(gt_target, bool):
            gt_target = "y" if gt_target else "n"
        npta = ManyArrays({"sourceval": sourceval})
        npts = len(npta)
        mpta = ManyArrays({"targval": targval})
        mpts = len(mpta)
        icount_interp = c_int()
        res = self.pestutils.interpolate_blend_using_file(
            byref(self.create_char_array(bytes(factorfile), "LENFILENAME")),
            byref(c_int(factorfiletype)),
            byref(c_int(npts)),
            byref(c_int(mpts)),
            byref(c_int(transtype)),
            byref(c_char(lt_target.encode())),
            byref(c_char(gt_target.encode())),
            npta.sourceval,
            mpta.targval,
            byref(icount_interp),
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("applied interpolation factors from %r", factorfile.name)
        return {
            "targval": mpts.targval.copy("A"),
            "icount_interp": icount_interp.value,
        }

    def ipd_interpolate_2d(
        self,
        # npts: int,  # determined from ecs.shape[0]
        ecs: npt.ArrayLike,
        ncs: npt.ArrayLike,
        zns: int | npt.ArrayLike,
        sourceval: npt.ArrayLike,
        # mpts: int,  # determined from ect.shape[0]
        ect: npt.ArrayLike,
        nct: npt.ArrayLike,
        znt: int | npt.ArrayLike,
        transtype: int | str | enum.TransType,
        anis: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        invpow: float | npt.ArrayLike,
    ) -> npt.NDArray[np.float64]:
        """Undertake 2D inverse-power-of-distance spatial interpolation.

        Parameters
        ----------
        ecs, ncs : array_like
            Source point coordinates, each 1D array with shape (npts,).
        zns : int or array_like
            Source point zones, integer or 1D array with shape (npts,).
        sourceval : array_like
            Source values, 1D array with shape (npts,).
        ect, nct : array_like
            Target point coordinates, each 1D array with shape (mpts,).
        znt : int or array_like
            Target point zones, integer or 1D array with shape (mpts,).
        transtype : int, str, enum.TransType
            Tranformation type, where 0 is none and 1 is log.
        anis : float or array_like
            Local anisotropy, float or 1D array with shape (mpts,).
        bearing : float or array_like
            Local anisotropy bearing, float or 1D array with shape (mpts,).
        invpow : float or array_like
            Local inverse power of distance, float or 1D array with shape (mpts,).

        Returns
        -------
        npt.NDArray[np.float64]
            Values calculated for targets.
        """
        npta = ManyArrays(
            {"ecs": ecs, "ncs": ncs, "sourceval": sourceval}, int_any={"zns": zns}
        )
        npts = len(npta)
        mpta = ManyArrays(
            {"ect": ect, "nct": nct},
            {"anis": anis, "bearing": bearing, "invpow": invpow},
            {"znt": znt},
        )
        mpts = len(mpta)
        if isinstance(transtype, str):
            transtype = enum.TransType.get_value(transtype)
        targval = np.zeros(mpts, np.float64, order="F")
        res = self.pestutils.ipd_interpolate_2d(
            byref(c_int(npts)),
            npta.ecs,
            npta.ncs,
            npta.zns,
            npta.sourceval,
            byref(c_int(mpts)),
            mpta.ect,
            mpta.nct,
            mpta.znt,
            targval,
            byref(c_int(transtype)),
            mpta.anis,
            mpta.bearing,
            mpta.invpow,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("undertook 2D inverse-power-of-distance spatial interpolation")
        return targval.copy("A")

    def ipd_interpolate_3d(
        self,
        # npts: int,  # determined from ecs.shape[0]
        ecs: npt.ArrayLike,
        ncs: npt.ArrayLike,
        zcs: npt.ArrayLike,
        zns: int | npt.ArrayLike,
        sourceval: npt.ArrayLike,
        # mpts: int,  # determined from ect.shape[0]
        ect: npt.ArrayLike,
        nct: npt.ArrayLike,
        zct: npt.ArrayLike,
        znt: int | npt.ArrayLike,
        transtype: int | str | enum.TransType,
        ahmax: float | npt.ArrayLike,
        ahmin: float | npt.ArrayLike,
        avert: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        dip: float | npt.ArrayLike,
        rake: float | npt.ArrayLike,
        invpow: float | npt.ArrayLike,
    ) -> npt.NDArray[np.float64]:
        """Undertake 3D inverse-power-of-distance spatial interpolation.

        Parameters
        ----------
        ecs, ncs, zcs : array_like
            Source point coordinates, each 1D array with shape (npts,).
        zns : int or array_like
            Source point zones, integer or 1D array with shape (npts,).
        sourceval : array_like
            Source values, 1D array with shape (npts,).
        ect, nct, zct : array_like
            Target point coordinates, each 1D array with shape (mpts,).
        znt : int or array_like
            Target point zones, integer or 1D array with shape (mpts,).
        transtype : int, str, enum.TransType
            Tranformation type, where 0 is none and 1 is log.
        ahmax, ahmin, avert : float or array_like
            Relative correlation lengths, float or 1D array with shape (mpts,).
        bearing, dip, rake : float or array_like
            Correlation directions, float or 1D array with shape (mpts,).
        invpow : float or array_like
            Local inverse power of distance, float or 1D array with shape (mpts,).

        Returns
        -------
        npt.NDArray[np.float64]
            Values calculated for targets.
        """
        npta = ManyArrays(
            {"ecs": ecs, "ncs": ncs, "zcs": zcs, "sourceval": sourceval},
            int_any={"zns": zns},
        )
        npts = len(npta)
        mpta = ManyArrays(
            {"ect": ect, "nct": nct, "zct": zct},
            {
                "ahmax": ahmax,
                "ahmin": ahmin,
                "avert": avert,
                "bearing": bearing,
                "dip": dip,
                "rake": rake,
                "invpow": invpow,
            },
            {"znt": znt},
        )
        mpts = len(mpta)
        if isinstance(transtype, str):
            transtype = enum.TransType.get_value(transtype)
        targval = np.zeros(mpts, np.float64, order="F")
        res = self.pestutils.ipd_interpolate_3d(
            byref(c_int(npts)),
            npta.ecs,
            npta.ncs,
            npta.zcs,
            npta.zns,
            npta.sourceval,
            byref(c_int(mpts)),
            mpta.ect,
            mpta.nct,
            mpta.zct,
            mpta.znt,
            targval,
            byref(c_int(transtype)),
            mpta.ahmax,
            mpta.ahmin,
            mpta.avert,
            mpta.bearing,
            mpta.dip,
            mpta.rake,
            mpta.invpow,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("undertook 3D inverse-power-of-distance spatial interpolation")
        return targval.copy("A")

    def initialize_randgen(self, iseed: int) -> None:
        """
        Initialize the random number generator.

        Parameters
        ----------
        iseed : int
            Seed value.
        """
        res = self.pestutils.initialize_randgen(byref(c_int(iseed)))
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("initialized the random number generator")

    def fieldgen2d_sva(
        self,
        # nnode: int,  # determined from ec.shape[0]
        ec: npt.ArrayLike,
        nc: npt.ArrayLike,
        area: float | npt.ArrayLike,
        active: int | npt.ArrayLike,
        mean: float | npt.ArrayLike,
        var: float | npt.ArrayLike,
        aa: float | npt.ArrayLike,
        anis: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        transtype: int | str | enum.TransType,
        avetype: int | str | enum.VarioType,
        power: float,
        # ldrand: int,  # same as nnode
        nreal: int,
    ) -> npt.NDArray[np.float64]:
        """
        Generate 2D stochastic fields based on a spatially varying variogram.

        Parameters
        ----------
        ec, nc : array_like
            Model grid coordinates, each 1D array with shape (nnode,).
        area : float or array_like
            Areas of grid cells.
        active : int or array_like
            Inactive grid cells are equal to zero.
        mean : float or array_like
            Mean value of stochastic field.
        var : float or array_like
            Variance of stochastic field.
        aa : float or array_like
            Averaging function spatial dimension.
        anis : float or array_like
            Anisotropy ratio.
        bearing : float or array_like
            Bearing of principal anisotropy axis.
        transtype : int, str or enum.TransType
            Stochastic field pertains to natural(0) or log(1) properties.
        avetype : int, str or enum.VarioType
            Averaging function type, where 1:spher, 2:exp, 3:gauss, 4:pow.
        power : float
            Power used if avetype is 4 (pow).
        nreal : int
            Number of realisations to generate.

        Returns
        -------
        npt.NDArray[np.float64]
            Realisations with shape (nnode, nreal).
        """
        node = ManyArrays(
            {"ec": ec, "nc": nc},
            {
                "area": area,
                "mean": mean,
                "var": var,
                "aa": aa,
                "anis": anis,
                "bearing": bearing,
            },
            {"active": active},
        )
        if isinstance(transtype, str):
            transtype = enum.TransType.get_value(transtype)
        if isinstance(avetype, str):
            avetype = enum.VarioType.get_value(avetype)
        ldrand = nnode = len(node)
        randfield = np.zeros((ldrand, nreal), np.float64, order="F")
        res = self.pestutils.fieldgen2d_sva(
            byref(c_int(nnode)),
            node.ec,
            node.nc,
            node.area,
            node.active,
            node.mean,
            node.var,
            node.aa,
            node.anis,
            node.bearing,
            byref(c_int(transtype)),
            byref(c_int(avetype)),
            byref(c_double(power)),
            byref(c_int(ldrand)),
            byref(c_int(nreal)),
            randfield,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("generated 2D stochastic fields for %d realisations", nreal)
        return randfield.copy("A")

    def fieldgen3d_sva(
        self,
        # nnode: int,  # determined from ec.shape[0]
        ec: npt.ArrayLike,
        nc: npt.ArrayLike,
        zc: npt.ArrayLike,
        area: float | npt.ArrayLike,
        height: float | npt.ArrayLike,
        active: int | npt.ArrayLike,
        mean: float | npt.ArrayLike,
        var: float | npt.ArrayLike,
        ahmax: float | npt.ArrayLike,
        ahmin: float | npt.ArrayLike,
        avert: float | npt.ArrayLike,
        bearing: float | npt.ArrayLike,
        dip: float | npt.ArrayLike,
        rake: float | npt.ArrayLike,
        transtype: int | str | enum.TransType,
        avetype: int | str | enum.VarioType,
        power: float,
        # ldrand: int,  # same as nnode
        nreal: int,
    ) -> npt.NDArray[np.float64]:
        """
        Generate 3D stochastic fields based on a spatially varying variogram.

        Parameters
        ----------
        ec, nc, nz : array_like
            Model grid coordinates, each 1D array with shape (nnode,).
        area, height : float or array_like
            Areas and height of grid cells.
        active : int or array_like
            Inactive grid cells are equal to zero.
        mean : float or array_like
            Mean value of stochastic field.
        var : float or array_like
            Variance of stochastic field.
        ahmax, ahmin, avert : float or array_like
            Averaging function correlation lengths.
        bearing : float or array_like
            Bearing of ahmax direction.
        dip : float or array_like
            Dip of ahmax direction.
        rake : float or array_like
            Rotation of ahmin direction.
        transtype : int, str or enum.TransType
            Stochastic field pertains to natural(0) or log(1) properties.
        avetype : int, str or enum.VarioType
            Averaging function type, where 1:spher, 2:exp, 3:gauss, 4:pow.
        power : float
            Power used if avetype is 4 (pow).
        nreal : int
            Number of realisations to generate.

        Returns
        -------
        npt.NDArray[np.float64]
            Realisations with shape (nnode, nreal).
        """
        node = ManyArrays(
            {"ec": ec, "nc": nc, "zc": zc},
            {
                "area": area,
                "height": height,
                "mean": mean,
                "var": var,
                "ahmax": ahmax,
                "ahmin": ahmin,
                "avert": avert,
                "bearing": bearing,
                "dip": dip,
                "rake": rake,
            },
            {"active": active},
        )
        if isinstance(transtype, str):
            transtype = enum.TransType.get_value(transtype)
        if isinstance(avetype, str):
            avetype = enum.VarioType.get_value(avetype)
        ldrand = nnode = len(node)
        randfield = np.zeros((ldrand, nreal), np.float64, order="F")
        res = self.pestutils.fieldgen3d_sva(
            byref(c_int(nnode)),
            node.ec,
            node.nc,
            node.zc,
            node.area,
            node.height,
            node.active,
            node.mean,
            node.var,
            node.ahmax,
            node.ahmin,
            node.avert,
            node.bearing,
            node.dip,
            node.rake,
            byref(c_int(transtype)),
            byref(c_int(avetype)),
            byref(c_double(power)),
            byref(c_int(ldrand)),
            byref(c_int(nreal)),
            randfield,
        )
        if res != 0:
            raise PestUtilsLibError(self.retrieve_error_message())
        self.logger.info("generated 3D stochastic fields for %d realisations", nreal)
        return randfield.copy("A")



"""Data module."""
from __future__ import annotations

from enum import Enum
from inspect import isclass
from typing import Any

import numpy as np
import numpy.typing as npt

from pypestutils.enum import ParamEnum

__all__ = ["ManyArrays", "validate_scalar"]


def validate_scalar(name: str, value: Any, **kwargs) -> None:
    """Validate scalar value according to supported kwargs.

    Parameters
    ----------
    name : str
        Name of parameter, used for error message.
    value : any
        Value of parameter.
    kwargs : dict
        Supported validation keywords are: isfinite, gt, ge, lt, le, isin,
        enum, minlen, maxlen and leneq.

    Raises
    ------
    ValueError
        When value fails validation criteria.
    TypeError
        When parameter use is not expected.
    NotImplementedError
        When keyword is not recognized.
    """
    if not np.isscalar(value):
        raise TypeError(f"'{name}' is not a scalar value")
    if "isfinite" in kwargs:
        if kwargs.pop("isfinite") is not True:
            raise TypeError("isfinite must be True")
        if not np.isfinite(value):
            raise ValueError(f"'{name}' must be finite (was {value!r})")
    if "gt" in kwargs:
        gt = kwargs.pop("gt")
        if not (value > gt):
            raise ValueError(f"'{name}' must be greater than {gt} (was {value!r})")
    if "ge" in kwargs:
        ge = kwargs.pop("ge")
        if not (value >= ge):
            raise ValueError(
                f"'{name}' must be greater than or equal to {ge} (was {value!r})"
            )
    if "lt" in kwargs:
        lt = kwargs.pop("lt")
        if not (value < lt):
            raise ValueError(f"'{name}' must be less than {lt} (was {value!r})")
    if "le" in kwargs:
        le = kwargs.pop("le")
        if not (value <= le):
            raise ValueError(
                f"'{name}' must be less than or equal to {le} (was {value!r})"
            )
    if "isin" in kwargs:
        isin = kwargs.pop("isin")
        if not np.isin(value, isin):
            raise ValueError(f"'{name}' must be in {isin} (was {value!r})")
    if "enum" in kwargs:
        enum_t = kwargs.pop("enum")
        if not (isclass(enum_t) and issubclass(enum_t, ParamEnum)):
            raise TypeError("enum must be a subclass of ParamEnum")
        elif isinstance(value, Enum) and not isinstance(value, enum_t):
            raise TypeError(f"'{value!s}' is not an enum {enum_t.__name__}")
        elif not isinstance(value, int):
            raise TypeError(f"enum value must be either {enum_t.__name__} or int")
        valid_options = enum_t.get_valid_options()
        if not np.isin(value, list(valid_options.keys())):
            enum_str = ", ".join([f"{v} ({n})" for (v, n) in valid_options.items()])
            raise ValueError(
                f"'{name}' must be in enum {enum_t.__name__} {enum_str} (was {value!r})"
            )
    if "minlen" in kwargs:
        valuelen = len(value)
        minlen = kwargs.pop("minlen")
        if minlen < 1:
            raise TypeError("minlen must be 1 or more")
        elif minlen == 1 and valuelen < 1:  # special case `minlen=1`
            raise ValueError(f"'{name}' cannot have zero len")
        elif valuelen < minlen:
            raise ValueError(f"'{name}' has a min len {minlen} (was {len(value)})")
    if "maxlen" in kwargs:
        valuelen = len(value)
        maxlen = kwargs.pop("maxlen")
        if valuelen > maxlen:
            raise ValueError(f"'{name}' has a max len {maxlen} (was {valuelen})")
    if "leneq" in kwargs:
        valuelen = len(value)
        leneq = kwargs.pop("leneq")
        if valuelen != leneq:
            raise ValueError(f"'{name}' must have len {leneq} (was {valuelen})")
    if kwargs:
        raise NotImplementedError(f"unhandled kwargs {kwargs}")


class ManyArrays:
    """Gather and check arrays and, if needed, fill-out scalars.

    All arrays are 1D with the same shape (or length). Float arrays are always
    float64, and integer arrays are always int32. All arrays are contiguous.
    This class is used as a pre-processor input for ctypes.

    Parameters
    ----------
    float_arrays : dict of array_like, optional
        Dict of 1D arrays, assume to have same shape.
    float_any : dict of array_like or float, optional
        Dict of float or 1D arrays.
    int_any : dict of array_like or int, optional
        Dict of int or 1D arrays.
    ar_len : int, optional
        If specified, this is used for the expected array size and shape.
    """

    shape = ()  # type: tuple | tuple[int]
    _names = []  # type: list[str]

    def __init__(
        self,
        float_arrays: dict[str, npt.ArrayLike] = {},
        float_any: dict[str, float | npt.ArrayLike] = {},
        int_any: dict[str, int | npt.ArrayLike] = {},
        ar_len: int | None = None,
    ) -> None:
        self._names = []
        # find common array size
        if ar_len is not None:
            if not isinstance(ar_len, int):
                raise TypeError("'ar_len' must be int")
            self.shape = (ar_len,)
        for name in float_arrays.keys():
            if name in self._names:
                raise KeyError(f"'{name}' defined more than once")
            self._names.append(name)
            # Each must be 1D and the same shape
            ar = np.array(float_arrays[name], np.float64, order="F", copy=False)
            if ar.ndim != 1:
                raise ValueError(f"expected '{name}' ndim to be 1; found {ar.ndim}")
            if not self.shape:
                self.shape = ar.shape
            elif ar.shape != self.shape:
                raise ValueError(
                    f"expected '{name}' shape to be {self.shape}; found {ar.shape}"
                )
            setattr(self, name, ar)
        for name in float_any.keys():
            if name in self._names:
                raise KeyError(f"'{name}' defined more than once")
            self._names.append(name)
            float_any[name] = ar = np.array(
                float_any[name], np.float64, order="F", copy=False
            )
            if not self.shape and ar.ndim == 1:
                self.shape = ar.shape
        for name in int_any.keys():
            if name in self._names:
                raise KeyError(f"'{name}' defined more than once")
            self._names.append(name)
            int_any[name] = ar = np.array(int_any[name], order="F", copy=False)
            if not self.shape and ar.ndim == 1:
                self.shape = ar.shape
        if not self.shape:
            self.shape = (1,)  # if all scalars, assume this size
        for name in float_any.keys():
            ar = float_any[name]
            if ar.ndim == 0:
                ar = np.full(self.shape, ar)
            elif ar.ndim != 1:
                raise ValueError(f"expected '{name}' ndim to be 1; found {ar.ndim}")
            elif ar.shape != self.shape:
                raise ValueError(
                    f"expected '{name}' shape to be {self.shape}; found {ar.shape}"
                )
            setattr(self, name, ar)
        for name in int_any.keys():
            ar = int_any[name]
            if ar.ndim == 0:
                ar = np.full(self.shape, ar)
            elif ar.ndim != 1:
                raise ValueError(f"expected '{name}' ndim to be 1; found {ar.ndim}")
            elif ar.shape != self.shape:
                raise ValueError(
                    f"expected '{name}' shape to be {self.shape}; found {ar.shape}"
                )
            if not np.issubdtype(ar.dtype, np.integer):
                raise ValueError(
                    f"expected '{name}' to be integer type; found {ar.dtype}"
                )
            setattr(self, name, ar.astype(np.int32, copy=False))

    def __len__(self) -> int:
        """Return length of dimension from shape[0]."""
        return self.shape[0]

    def validate(self, name, **kwargs) -> None:
        """Validate array values.

        Parameters
        ----------
        name : str
            Name of parameter.
        kwargs : dict
            Supported validation keywords are: isfinite, gt, ge, lt, le, isin
            and enum.

        Raises
        ------
        ValueError
            When 1 or more array elements fail validation criteria.
        TypeError
            When parameter use is not expected.
        NotImplementedError
            When keyword is not recognized.
        """
        if name not in self._names:
            raise KeyError(f"'{name}' not found")
        ar = getattr(self, name)
        dtp = ar.dtype
        typ = dtp.type
        if "isfinite" in kwargs:
            if kwargs.pop("isfinite") is not True:
                raise TypeError("isfinite must be True")
            if not np.isfinite(ar).all():
                raise ValueError(f"'{name}' must be finite")
        if "gt" in kwargs:
            gt = typ(kwargs.pop("gt"))
            if not (ar > gt).all():
                raise ValueError(f"'{name}' must be greater than {gt}")
        if "ge" in kwargs:
            ge = typ(kwargs.pop("ge"))
            if not (ar >= ge).all():
                raise ValueError(f"'{name}' must be greater than or equal to {ge}")
        if "lt" in kwargs:
            lt = typ(kwargs.pop("lt"))
            if not (ar < lt).all():
                raise ValueError(f"'{name}' must be less than {lt}")
        if "le" in kwargs:
            le = typ(kwargs.pop("le"))
            if not (ar <= le).all():
                raise ValueError(f"'{name}' must be less than or equal to {le}")
        if "isin" in kwargs:
            isin = kwargs.pop("isin")
            if not np.isin(ar, isin).all():
                raise ValueError(f"'{name}' must be in {isin}")
        if "enum" in kwargs:
            enum_t = kwargs.pop("enum")
            if not (isclass(enum_t) and issubclass(enum_t, ParamEnum)):
                raise TypeError("enum must be a subclass of ParamEnum")
            elif not np.issubdtype(dtp, np.integer):
                raise TypeError(f"'{name}' values must be integer type")
            valid_options = enum_t.get_valid_options()
            if not np.isin(ar, list(valid_options.keys())).all():
                enum_str = ", ".join([f"{v} ({n})" for (v, n) in valid_options.items()])
                raise ValueError(
                    f"'{name}' must be in enum {enum_t.__name__} {enum_str}"
                )
        if kwargs:
            raise NotImplementedError(f"unhandled kwargs {kwargs}")



"""Locate pestutils shared library by any means necessary."""
import ctypes
import os
import platform
import sys
from ctypes.util import find_library
from pathlib import Path

# the current working directory of this file
pkg_dir = Path(__file__).parent

# generate a bunch of candidate locations where the
# shared library *might* be hanging out
_candidates = [
    os.environ.get("PESTUTILS_LIBRARY", None),
    str(pkg_dir / "lib"),  # see scripts/build_lib.sh
    str(pkg_dir.parent / "inst" / ("bin" if os.name == "nt" else "lib")),
    ".",
]


def load() -> ctypes.CDLL:
    """Load the pestutils shared library.

    Returns
    -------
    ctypes.CDLL
        Loaded shared library
    """
    if os.name == "nt":
        lib_name = "pestutils.dll"

        # add search paths for conda installs
        if (
            os.path.exists(os.path.join(sys.prefix, "conda-meta"))
            or "conda" in sys.version
        ):
            _candidates.append(os.path.join(sys.prefix, "Library", "bin"))

        # get the current PATH
        oldenv = os.environ.get("PATH", "").strip().rstrip(";")
        # run through our list of candidate locations
        for path in _candidates:
            if not path or not os.path.exists(path):
                continue
            # temporarily add the path to the PATH environment variable
            # so Windows can find additional DLL dependencies.
            os.environ["PATH"] = ";".join([path, oldenv])
            try:
                rt = ctypes.cdll.LoadLibrary(os.path.join(path, lib_name))
                if rt is not None:
                    return rt
            except OSError:
                pass
            except BaseException as err:
                print(f"pypestutils.finder unexpected error: {err!s}")
            finally:
                os.environ["PATH"] = oldenv
        raise OSError(f"could not find or load {lib_name}")

    elif os.name == "posix":
        # posix includes both mac and linux
        # use the extension for the specific platform
        if platform.system() == "Darwin":
            # macos shared libraries are `.dylib`
            lib_name = "libpestutils.dylib"
        else:
            # linux shared libraries are `.so`
            lib_name = "libpestutils.so"

        # get the starting working directory
        cwd = os.getcwd()
        for cand in _candidates:
            if cand is None:
                continue
            elif os.path.isdir(cand):
                # if our candidate is a directory use best guess
                path = cand
                target = os.path.join(cand, lib_name)
            elif os.path.isfile(cand):
                # if candidate is just a file use that
                path = os.path.split(cand)[0]
                target = cand
            else:
                continue

            if not os.path.exists(target):
                continue

            try:
                # move to the location we're checking
                os.chdir(path)
                # try loading the target file candidate
                rt = ctypes.cdll.LoadLibrary(target)
                if rt is not None:
                    return rt
            except BaseException as err:
                print(f"pypestutils.finder ({target}) unexpected error: {err!s}")
            finally:
                os.chdir(cwd)

    try:
        # try loading library using LD path search
        path = find_library("pestutils")
        if path is not None:
            return ctypes.cdll.LoadLibrary(path)

    except BaseException:
        pass

    raise OSError("Could not load pestutils library")



"""Low-level Fortran-Python ctypes functions."""
from __future__ import annotations

from ctypes import ARRAY, CDLL, POINTER, c_char, c_double, c_int

from numpy.ctypeslib import ndpointer

# Cache variables by uppercase dimvar name
_dimvar_cache = {}
_char_array_cache = {}
# other lengths not defined in dimvar
_misc_lengths = {
    "LENVARTYPE": 17,
    "LENFLOWTYPE": 17,
}


def get_dimvar_int(lib: CDLL, name: str) -> int:
    """Get dimvar constant integer from library instance.

    Parameters
    ----------
    lib : CDLL
        Ctypes library instance.
    name : str
        Uppercase name of variable in dimvar or other custom name.

    Returns
    -------
    int

    Raises
    ------
    ValueError
        If name is not defined in lib object.
    """
    if name in _dimvar_cache:
        return _dimvar_cache[name]
    elif name in _misc_lengths:
        # Special consideration for constants not specified by dimvar
        return _misc_lengths[name]
    c_var = c_int.in_dll(lib, name)
    _dimvar_cache[name] = c_var.value
    return c_var.value


def get_char_array(lib: CDLL, name: str):
    """Get c_char Array with a fixed size from dimvar.

    Parameters
    ----------
    lib : CDLL
        Ctypes library instance.
    name : str
        Uppercase name of variable in dimvar or other custom name.
    """
    if name in _char_array_cache:
        return _char_array_cache[name]
    size = get_dimvar_int(lib, name)
    array_type = ARRAY(c_char, size)
    _char_array_cache[name] = array_type
    return array_type


def prototype(lib) -> None:
    """Add ctypes prototypes for each function in pestutils.

    Parameters
    ----------
    lib : CDLL
        Ctypes library instance, which is modified in-place
    """
    # Generate c_char Array types based on dimvar sizes
    filename_t = get_char_array(lib, "LENFILENAME")
    message_t = get_char_array(lib, "LENMESSAGE")
    gridname_t = get_char_array(lib, "LENGRIDNAME")
    vartype_t = get_char_array(lib, "LENVARTYPE")
    flowtype_t = get_char_array(lib, "LENFLOWTYPE")

    # inquire_modflow_binary_file_specs(
    #   filein,fileout,isim,itype,iprec,narray,ntime)
    lib.inquire_modflow_binary_file_specs.argtypes = (
        POINTER(filename_t),  # filein, in
        POINTER(filename_t),  # fileout, in
        POINTER(c_int),  # isim, in
        POINTER(c_int),  # itype, in
        POINTER(c_int),  # iprec, out
        POINTER(c_int),  # narray, out
        POINTER(c_int),  # ntime, out
    )
    lib.inquire_modflow_binary_file_specs.restype = c_int

    # retrieve_error_message(errormessage)
    lib.retrieve_error_message.argtypes = (POINTER(message_t),)  # errormessage, out
    lib.retrieve_error_message.restype = c_int

    # install_structured_grid(
    #   gridname,ncol,nrow,nlay,icorner,e0,n0,rotation,delr,delc)
    lib.install_structured_grid.argtypes = (
        POINTER(gridname_t),  # gridname, in
        POINTER(c_int),  # ncol, in
        POINTER(c_int),  # nrow, in
        POINTER(c_int),  # nlay, in
        POINTER(c_int),  # icorner, in
        POINTER(c_double),  # e0, in
        POINTER(c_double),  # n0, in
        POINTER(c_double),  # rotation, in
        ndpointer(c_double, ndim=1, flags="F"),  # delr(ncol), in
        ndpointer(c_double, ndim=1, flags="F"),  # delc(nrow), in
    )
    lib.install_structured_grid.restype = c_int

    # get_cell_centres_structured(gridname,ncpl,cellx,celly)
    lib.get_cell_centres_structured.argtypes = (
        POINTER(gridname_t),  # gridname, in
        POINTER(c_int),  # ncpl, in
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # cellx(ncells), out
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # celly(ncells), out
    )
    lib.get_cell_centres_structured.restype = c_int

    # uninstall_structured_grid(gridname)
    lib.uninstall_structured_grid.argtypes = (POINTER(gridname_t),)  # gridname, in
    lib.uninstall_structured_grid.restype = c_int

    # free_all_memory()
    lib.free_all_memory.argtypes = ()
    lib.free_all_memory.restype = c_int

    # interp_from_structured_grid(
    #   gridname,depvarfile,isim,iprec,ntime,vartype,interpthresh,nointerpval,
    #   npts,ecoord,ncoord,layer,nproctime,simtime,simstate)
    lib.interp_from_structured_grid.argtypes = (
        POINTER(gridname_t),  # gridname, in
        POINTER(filename_t),  # depvarfile, in
        POINTER(c_int),  # isim, in
        POINTER(c_int),  # iprec, in
        POINTER(c_int),  # ntime, in
        POINTER(vartype_t),  # vartype, in
        POINTER(c_double),  # interpthresh, in
        POINTER(c_double),  # nointerpval, in
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ecoord, in
        ndpointer(c_double, ndim=1, flags="F"),  # ncoord, in
        ndpointer(c_int, ndim=1, flags="F"),  # layer, in
        POINTER(c_int),  # nproctime, out
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # simtime(ntime), out
        ndpointer(c_double, ndim=2, flags=("F", "W")),  # simstate(ntime,npts), out
    )
    lib.interp_from_structured_grid.restype = c_int

    # interp_to_obstime(
    #   nsimtime,nproctime,npts,simtime,simval,interpthresh,how_extrap,
    #   time_extrap,nointerpval,nobs,obspoint,obstime,obssimval)
    lib.interp_to_obstime.argtypes = (
        POINTER(c_int),  # nsimtime, in
        POINTER(c_int),  # nproctime, in
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # simtime(nsimtime), in
        ndpointer(c_double, ndim=2, flags="F"),  # simval(nsimtime,npts), in
        POINTER(c_double),  # interpthresh, in
        POINTER(c_char),  # how_extrap, in
        POINTER(c_double),  # time_extrap, in
        POINTER(c_double),  # nointerpval, in
        POINTER(c_int),  # nobs, in
        ndpointer(c_int, ndim=1, flags="F"),  # obspoint(nobs), in
        ndpointer(c_double, ndim=1, flags="F"),  # obstime(nobs), in
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # obssimval(nobs), out
    )
    lib.interp_to_obstime.restype = c_int

    # install_mf6_grid_from_file(
    #   gridname,grbfile,idis,ncells,ndim1,ndim2,ndim3)
    lib.install_mf6_grid_from_file.argtypes = (
        POINTER(gridname_t),  # gridname, in
        POINTER(filename_t),  # grbfile, in
        POINTER(c_int),  # idis, out
        POINTER(c_int),  # ncells, out
        POINTER(c_int),  # ndim1, out
        POINTER(c_int),  # ndim2, out
        POINTER(c_int),  # ndim3, out
    )
    lib.install_mf6_grid_from_file.restype = c_int

    # get_cell_centres_mf6(gridname,ncells,cellx,celly,cellz)
    lib.get_cell_centres_mf6.argtypes = (
        POINTER(gridname_t),  # gridname, in
        POINTER(c_int),  # ncells, in
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # cellx(ncells), out
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # celly(ncells), out
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # cellz(ncells), out
    )
    lib.get_cell_centres_mf6.restype = c_int

    # uninstall_mf6_grid(gridname)
    lib.uninstall_mf6_grid.argtypes = (POINTER(gridname_t),)  # gridname, in
    lib.uninstall_mf6_grid.restype = c_int

    # calc_mf6_interp_factors(
    #   gridname,npts,ecoord,ncoord,layer,factorfile,
    #   factorfiletype,blnfile,interp_success)
    lib.calc_mf6_interp_factors.argtypes = (
        POINTER(gridname_t),  # gridname, in
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ecoord, in
        ndpointer(c_double, ndim=1, flags="F"),  # ncoord, in
        ndpointer(c_int, ndim=1, flags="F"),  # layer, in
        POINTER(filename_t),  # factorfile, in
        POINTER(c_int),  # factorfiletype, in
        POINTER(filename_t),  # blnfile, in
        ndpointer(c_int, ndim=1, flags=("F", "W")),  # interp_success, out
    )
    lib.calc_mf6_interp_factors.restype = c_int

    # interp_from_mf6_depvar_file(
    #   depvarfile,factorfile,factorfiletype,ntime,vartype,interpthresh,
    #   reapportion,nointerpval,npts,nproctime,simtime,simstate)
    lib.interp_from_mf6_depvar_file.argtypes = (
        POINTER(filename_t),  # depvarfile, in
        POINTER(filename_t),  # factorfile, in
        POINTER(c_int),  # factorfiletype, in
        POINTER(c_int),  # ntime, in
        POINTER(vartype_t),  # vartype(17), in
        POINTER(c_double),  # interpthresh, in
        POINTER(c_int),  # reapportion, in
        POINTER(c_double),  # nointerpval, in
        POINTER(c_int),  # npts, in
        POINTER(c_int),  # nproctime, out
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # simtime(ntime), out
        ndpointer(c_double, ndim=2, flags=("F", "W")),  # simstate(ntime,npts), out
    )
    lib.interp_from_mf6_depvar_file.restype = c_int

    # extract_flows_from_cbc_file(
    #   cbcfile,flowtype,isim,iprec,ncell,izone,nzone,numzone,zonenumber,
    #   ntime,nproctime,timestep,stressperiod,simtime,simflow)
    lib.extract_flows_from_cbc_file.argtypes = (
        POINTER(filename_t),  # cbcfile, in
        POINTER(flowtype_t),  # flowtype, in
        POINTER(c_int),  # isim, in
        POINTER(c_int),  # iprec, in
        POINTER(c_int),  # ncell, in
        ndpointer(c_int, ndim=1, flags="F"),  # izone(ncell), in
        POINTER(c_int),  # nzone, in
        POINTER(c_int),  # numzone, out
        ndpointer(c_int, ndim=1, flags=("F", "W")),  # zonenumber(nzone), out
        POINTER(c_int),  # ntime, in
        POINTER(c_int),  # nproctime, out
        ndpointer(c_int, ndim=1, flags=("F", "W")),  # timestep(ntime), out
        ndpointer(c_int, ndim=1, flags=("F", "W")),  # stressperiod(ntime), out
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # simtime(ntime), out
        ndpointer(c_double, ndim=2, flags=("F", "W")),  # simflow(ntime,nzone), out
    )
    lib.extract_flows_from_cbc_file.restype = c_int

    # calc_kriging_factors_2d(
    #   npts,ecs,ncs,zns,mpts,ect,nct,znt,vartype,krigtype,aa,anis,bearing,
    #   searchrad,maxpts,minpts,factorfile,factorfiletype,icount_interp)
    lib.calc_kriging_factors_2d.argtypes = (
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ecs(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ncs(npts), in
        ndpointer(c_int, ndim=1, flags="F"),  # zns(npts), in
        POINTER(c_int),  # mpts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ect(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # nct(mpts), in
        ndpointer(c_int, ndim=1, flags="F"),  # znt(mpts), in
        POINTER(c_int),  # vartype, in
        POINTER(c_int),  # krigtype, in
        ndpointer(c_double, ndim=1, flags="F"),  # aa(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # anis(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(mpts), in
        POINTER(c_double),  # searchrad, in
        POINTER(c_int),  # maxpts, in
        POINTER(c_int),  # minpts, in
        POINTER(filename_t),  # factorfile, in
        POINTER(c_int),  # factorfiletype, in
        POINTER(c_int),  # icount_interp, out
    )
    lib.calc_kriging_factors_2d.restype = c_int

    # calc_kriging_factors_auto_2d(
    #   npts,ecs,ncs,zns,mpts,ect,nct,znt,krigtype,anis,bearing,
    #   factorfile,factorfiletype,icount_interp)
    lib.calc_kriging_factors_auto_2d.argtypes = (
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ecs(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ncs(npts), in
        ndpointer(c_int, ndim=1, flags="F"),  # zns(npts), in
        POINTER(c_int),  # mpts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ect(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # nct(mpts), in
        ndpointer(c_int, ndim=1, flags="F"),  # znt(mpts), in
        POINTER(c_int),  # krigtype, in
        ndpointer(c_double, ndim=1, flags="F"),  # anis(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(mpts), in
        POINTER(filename_t),  # factorfile, in
        POINTER(c_int),  # factorfiletype, in
        POINTER(c_int),  # icount_interp, out
    )
    lib.calc_kriging_factors_auto_2d.restype = c_int

    # calc_kriging_factors_3d(
    #   npts,ecs,ncs,zcs,zns,mpts,ect,nct,zct,znt,krigtype,nzone,zonenum,
    #   vartype,ahmax,ahmin,avert,bearing,dip,rake,srhmax,srhmin,srvert,
    #   maxpts,minpts,factorfile,factorfiletype,icount_interp)
    lib.calc_kriging_factors_3d.argtypes = (
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ecs(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ncs(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # zcs(npts), in
        ndpointer(c_int, ndim=1, flags="F"),  # zns(npts), in
        POINTER(c_int),  # mpts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ect(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # nct(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # zct(mpts), in
        ndpointer(c_int, ndim=1, flags="F"),  # znt(mpts), in
        POINTER(c_int),  # krigtype, in
        POINTER(c_int),  # nzone, in
        ndpointer(c_int, ndim=1, flags="F"),  # zonenum(nzone), in
        ndpointer(c_int, ndim=1, flags="F"),  # vartype(nzone), in
        ndpointer(c_double, ndim=1, flags="F"),  # ahmax(nzone), in
        ndpointer(c_double, ndim=1, flags="F"),  # ahmin(nzone), in
        ndpointer(c_double, ndim=1, flags="F"),  # avert(nzone), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(nzone), in
        ndpointer(c_double, ndim=1, flags="F"),  # dip(nzone), in
        ndpointer(c_double, ndim=1, flags="F"),  # rake(nzone), in
        POINTER(c_double),  # srhmax, in
        POINTER(c_double),  # srhmin, in
        POINTER(c_double),  # srvert, in
        POINTER(c_int),  # maxpts, in
        POINTER(c_int),  # minpts, in
        POINTER(filename_t),  # factorfile, in
        POINTER(c_int),  # factorfiletype, in
        POINTER(c_int),  # icount_interp, out
    )
    lib.calc_kriging_factors_3d.restype = c_int

    # krige_using_file(
    #   factorfile,factorfiletype,npts,mpts,krigtype,transtype,
    #   sourceval,targval,icount_interp,meanval)
    lib.krige_using_file.argtypes = (
        POINTER(filename_t),  # factorfile, in
        POINTER(c_int),  # factorfiletype, in
        POINTER(c_int),  # npts, in
        POINTER(c_int),  # mpts, in
        POINTER(c_int),  # krigtype, in
        POINTER(c_int),  # transtype, in
        ndpointer(c_double, ndim=1, flags="F"),  # sourceval(npts), in
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # targval(mpts), out
        POINTER(c_int),  # icount_interp, out
        ndpointer(c_double, ndim=1, flags="F"),  # meanval(mpts), in, optional
    )
    lib.krige_using_file.restype = c_int

    # build_covar_matrix_2d(
    #   npts,ec,nc,zn,vartype,nugget,aa,sill,anis,bearing,ldcovmat,covmat)
    lib.build_covar_matrix_2d.argtypes = (
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ec(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # nc(npts), in
        ndpointer(c_int, ndim=1, flags="F"),  # zn(npts), in
        POINTER(c_int),  # vartype, in
        ndpointer(c_double, ndim=1, flags="F"),  # nugget(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # aa(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # sill(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # anis(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(npts), in
        POINTER(c_int),  # ldcovmat, in
        ndpointer(c_double, ndim=2, flags=("F", "W")),  # covmat(ldcovmat,npts), out
    )
    lib.build_covar_matrix_2d.restype = c_int

    # build_covar_matrix_3d(
    #   npts,ec,nc,zc,zn,vartype,
    # nugget,sill,ahmax,ahmin,avert,bearing,dip,rake,ldcovmat,covmat)
    lib.build_covar_matrix_3d.argtypes = (
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ec(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # nc(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # zc(npts), in
        ndpointer(c_int, ndim=1, flags="F"),  # zn(npts), in
        POINTER(c_int),  # vartype, in
        ndpointer(c_double, ndim=1, flags="F"),  # nugget(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # sill(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ahmax(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ahmin(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # avert(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # dip(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # rake(npts), in
        POINTER(c_int),  # ldcovmat, in
        ndpointer(c_double, ndim=2, flags=("F", "W")),  # covmat(ldcovmat,npts), out
    )
    lib.build_covar_matrix_3d.restype = c_int

    # calc_structural_overlay_factors(
    #   npts,ecs,ncs,ids,conwidth,aa,structype,inverse_power,
    #   mpts,ect,nct,active,factorfile,factorfiletype,icount_interp)
    lib.calc_structural_overlay_factors.argtypes = (
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ecs(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ncs(npts), in
        ndpointer(c_int, ndim=1, flags="F"),  # ids(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # conwidth(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # aa(npts), in
        POINTER(c_int),  # structype, in
        POINTER(c_double),  # inverse_power, in
        POINTER(c_int),  # mpts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ect(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # nct(mpts), in
        ndpointer(c_int, ndim=1, flags="F"),  # active(mpts), in
        POINTER(filename_t),  # factorfile, in
        POINTER(c_int),  # factorfiletype, in
        POINTER(c_int),  # icount_interp, out
    )
    lib.calc_structural_overlay_factors.restype = c_int

    # interpolate_blend_using_file(
    #   factorfile,factorfiletype,npts,mpts,transtype,
    #   lt_target,gt_target,sourceval,targval,icount_interp)
    lib.interpolate_blend_using_file.argtypes = (
        POINTER(filename_t),  # factorfile, in
        POINTER(c_int),  # factorfiletype, in
        POINTER(c_int),  # npts, in
        POINTER(c_int),  # mpts, in
        POINTER(c_int),  # transtype, in
        POINTER(c_char),  # lt_target, in
        POINTER(c_char),  # gt_target, in
        ndpointer(c_double, ndim=1, flags="F"),  # sourceval(npts), in
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # targval(mpts), inout
        POINTER(c_int),  # icount_interp, out
    )
    lib.interpolate_blend_using_file.restype = c_int

    # ipd_interpolate_2d(npts,ecs,ncs,zns,sourceval,
    #   mpts,ect,nct,znt,targval,transtype,anis,bearing,invpow)
    lib.ipd_interpolate_2d.argtypes = (
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ecs(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ncs(npts), in
        ndpointer(c_int, ndim=1, flags="F"),  # zns(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # sourceval(npts), in
        POINTER(c_int),  # mpts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ect(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # nct(mpts), in
        ndpointer(c_int, ndim=1, flags="F"),  # znt(mpts), in
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # targval(mpts), out
        POINTER(c_int),  # transtype, in
        ndpointer(c_double, ndim=1, flags="F"),  # anis(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # invpow(mpts), in
    )
    lib.ipd_interpolate_2d.restype = c_int

    # ipd_interpolate_3d(npts,ecs,ncs,zcs,zns,sourceval,mpts,ect,nct,zct,znt,
    #   targval,transtype,ahmax,ahmin,avert,bearing,dip,rake,invpow)
    lib.ipd_interpolate_3d.argtypes = (
        POINTER(c_int),  # npts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ecs(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ncs(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # zcs(npts), in
        ndpointer(c_int, ndim=1, flags="F"),  # zns(npts), in
        ndpointer(c_double, ndim=1, flags="F"),  # sourceval(npts), in
        POINTER(c_int),  # mpts, in
        ndpointer(c_double, ndim=1, flags="F"),  # ect(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # nct(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # zct(mpts), in
        ndpointer(c_int, ndim=1, flags="F"),  # znt(mpts), in
        ndpointer(c_double, ndim=1, flags=("F", "W")),  # targval(mpts), out
        POINTER(c_int),  # transtype, in
        ndpointer(c_double, ndim=1, flags="F"),  # ahmax(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # ahmin(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # avert(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # dip(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # rake(mpts), in
        ndpointer(c_double, ndim=1, flags="F"),  # invpow(mpts), in
    )
    lib.ipd_interpolate_3d.restype = c_int

    # initialize_randgen(iseed)
    lib.initialize_randgen.argtypes = (POINTER(c_int),)  # iseed, in
    lib.initialize_randgen.restype = c_int

    # fieldgen2d_sva(
    #   nnode,ec,nc,area,active,mean,var,aa,anis,bearing,
    #   transtype,avetype,power,ldrand,nreal,randfield)
    lib.fieldgen2d_sva.argtypes = (
        POINTER(c_int),  # nnode, in
        ndpointer(c_double, ndim=1, flags="F"),  # ec(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # nc(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # area(nnode), in
        ndpointer(c_int, ndim=1, flags="F"),  # active(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # mean(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # var(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # aa(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # anis(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(nnode), in
        POINTER(c_int),  # transtype, in
        POINTER(c_int),  # avetype, in
        POINTER(c_double),  # power, in
        POINTER(c_int),  # ldrand, in
        POINTER(c_int),  # nreal, in
        ndpointer(c_double, ndim=2, flags=("F", "W")),  # randfield(ldrand,nreal), out
    )
    lib.fieldgen2d_sva.restype = c_int

    # fieldgen3d_sva(
    #   nnode,ec,nc,zc,area,height,active,mean,var,ahmax,ahmin,avert,
    #   bearing,dip,rake,transtype,avetype,power,ldrand,nreal,randfield)
    lib.fieldgen3d_sva.argtypes = (
        POINTER(c_int),  # nnode, in
        ndpointer(c_double, ndim=1, flags="F"),  # ec(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # nc(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # zc(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # area(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # height(nnode), in
        ndpointer(c_int, ndim=1, flags="F"),  # active(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # mean(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # var(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # ahmax(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # ahmin(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # avert(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # bearing(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # dip(nnode), in
        ndpointer(c_double, ndim=1, flags="F"),  # rake(nnode), in
        POINTER(c_int),  # transtype, in
        POINTER(c_int),  # avetype, in
        POINTER(c_double),  # power, in
        POINTER(c_int),  # ldrand, in
        POINTER(c_int),  # nreal, in
        ndpointer(c_double, ndim=2, flags=("F", "W")),  # randfield(ldrand,nreal), out
    )
    lib.fieldgen3d_sva.restype = c_int



"""Logger module."""
from __future__ import annotations

__all__ = ["get_logger"]

import logging
import sys
from logging import Logger


def get_logger(name: str, level: str | int = 0) -> Logger:
    """Return a named logger.

    Parameters
    ----------
    name : str
        Logger name.
    logger_level : str, int, optional
        Logger level, default 0 ("NOTSET"). Accepted values are
        "DEBUG" (10), "INFO" (20), "WARNING" (30), "ERROR" (40) or
        "CRITICAL" (50).

    Returns
    -------
    Logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)

    if not logger.hasHandlers():
        formatter = logging.Formatter("%(levelname)s:%(name)s: %(message)s")
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger



#!/usr/bin/env python3
import os

from setuptools import setup
from setuptools.command.install import install
from setuptools.dist import Distribution
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

# current working directory of this setup.py file
_cwd = os.path.abspath(os.path.split(__file__)[0])


class bdist_wheel(_bdist_wheel):
    def finalize_options(self):
        _bdist_wheel.finalize_options(self)
        self.root_is_pure = False


class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""

    def has_ext_modules(foo):
        return True


class InstallPlatlib(install):
    def finalize_options(self):
        """
        Copy the shared libraries into the wheel. Note that this
        will *only* check in `pypestutils/lib` rather than anywhere on
        the system so if you are building a wheel you *must* copy or
        symlink the `.so`/`.dll`/`.dylib` files into `pypestutils/lib`.
        """
        # use for checking extension types
        from fnmatch import fnmatch

        install.finalize_options(self)
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib
        # get the location of the shared library on the filesystem

        # where we're putting the shared library in the build directory
        target_dir = os.path.join(self.build_lib, "pypestutils", "lib")
        # where are we checking for shared libraries
        source_dir = os.path.join(_cwd, "pypestutils", "lib")

        # what patterns represent shared libraries
        patterns = {"*.so", "libpestutils*dylib", "*.dll"}

        if not os.path.isdir(source_dir):
            # no copying of binary parts to library
            # this is so `pip install .` works even
            # if `pypestutils/lib` isn't populated
            return

        for file_name in os.listdir(source_dir):
            # make sure file name is lower case
            check = file_name.lower()
            # use filename pattern matching to see if it is
            # a shared library format file
            if not any(fnmatch(check, p) for p in patterns):
                continue

            # if the source isn't a file skip it
            if not os.path.isfile(os.path.join(source_dir, file_name)):
                continue

            # make build directory if it doesn't exist yet
            if not os.path.isdir(target_dir):
                os.makedirs(target_dir)

            # copy the source file to the target directory
            self.copy_file(
                os.path.join(source_dir, file_name), os.path.join(target_dir, file_name)
            )


# See pyproject.toml for other project metadata
setup(
    name="pypestutils",
    distclass=BinaryDistribution,
    cmdclass={"bdist_wheel": bdist_wheel, "install": InstallPlatlib},
)



from __future__ import annotations

import os
import numpy as np
import pandas as pd

from .pestutilslib import PestUtilsLib


def mod2obs_mf6(gridinfo_fname: str,depvar_fname: str,obscsv_fname: str ,model_type: int,start_datetime: str | pd.TimeStamp,depvar_ftype=1,
                depvar_name="head",interp_thresh=1.0e+30,no_interp_val=1.0e+30,model_timeunit="d",
                time_extrap=1.0)->dict:

    """python implementation of mod2smp and mod2obs using modflow6 binary grid files
    Parameters
    ----------
    gridinfo_fname: str
        grid information file
    depvar_fname: str
        MODFLOW-6 output binary file
    obscsv_fname: str | pd.DataFrame
        observation information.  Must contain columns "site","x","y","datetime",and "layer"
    model_type: int
        type of model.  Must be either 31 (dis mf6) or 32 (disv mf6)
    start_datetime: str | datetime
        the simulation start datetime
    depvar_ftype : int
        the modflow-6 output file type.  1 for states, 2 or cell-by-cell budgets
    depvar_name: str
        the name of the dependent variable in `depvar_fname` to extract (for example "head")
    interp_thresh: float
        the upper limit above which extracted values are treated as invalid.  Default is 1.0+30
    no_interp_val: float
        value used to fill invalid/null extracted/interpolated values
    model_time_unit: str
        pandas style time unit.  Default is "d"ay
    time_extrap: float
        length of time units to extrapolate.  Default is 1.0 time unit

    Returns
    -------
    all_results: pd.DataFrame
        all simulated times at observation locations (ie mod2smp)
    interpolated_results: pd.DataFrame
        temporally interpolated simulated results at observation locations (ie mod2obs)
    """

    for fname in [gridinfo_fname,depvar_fname]:
        assert os.path.exists(fname),"file {0} not found".format(fname)
    lib = PestUtilsLib()
    is_mf6 = False
    is_structured = True
    model_type = int(model_type)
    if model_type == 1:
        is_mf6 = False
    elif model_type == 21:
        pass
    elif model_type == 22:
        is_structured = False
    elif model_type == 31:
        is_mf6 = True
    elif model_type == 32:
        is_mf6 = True
        is_structured = False
    elif model_type == 33:
        is_mf6 = True
        is_structured = False
    else:
        raise Exception("unrecognized 'model_type':{0}".format(model_type))

    depvar_ftype = int(depvar_ftype)
    if depvar_ftype not in [1,2]:
        raise Exception("unrecognized 'depvar_ftype':{0}".format(depvar_ftype))

    if is_mf6:
        grid_info = lib.install_mf6_grid_from_file("grid",gridinfo_fname)
    else:
        raise NotImplementedError()
    
    if isinstance(start_datetime,str):
        start_datetime = pd.to_datetime(start_datetime)

    depvar_info = lib.inquire_modflow_binary_file_specs(depvar_fname,depvar_fname+".out.csv",model_type,depvar_ftype)    
    depvar_df = pd.read_csv(depvar_fname+".out.csv")
    depvar_df.columns = [c.lower() for c in depvar_df.columns]
    #print(depvar_df)

    if isinstance(obscsv_fname,str):
        if not os.path.exists(obscsv_fname):
            raise Exception("obscsv_fname '{0}' not found".format(obscsv_fname))
        # todo: think about supporting a site sample file maybe?
        obsdf = pd.read_csv(os.path.join(obscsv_fname),parse_dates=["datetime"])
    elif isinstance(obscsv_fname,pd.DataFrame):
        obsdf = obscsv_fname.copy()
    else:
        raise Exception("obscsv arg type not recognized (looking for str or pd.DataFrame):'{0}'".format(type(obscsv_fname)))
    #check obsdf
    obsdf.columns = [c.lower() for c in obsdf.columns]
    for req_col in ["site","x","y","datetime","layer"]:
        if req_col not in obsdf.columns:
            raise Exception("observation dataframe missing column '{0}'".format(req_col))
    usitedf = obsdf.groupby("site").first()
    pth = os.path.split(depvar_fname)[0]
    fac_file = os.path.join(pth,"obs_interp_fac.bin")
    bln_file = fac_file.replace(".bin",".bln")
    interp_fac_results = lib.calc_mf6_interp_factors("grid",usitedf.x.values,usitedf.y.values,usitedf.layer.values,fac_file,"binary",bln_file)
    if 0 in interp_fac_results:
        print("warning: the following site(s) failed to have interpolation factors calculated:")
        fsites = usitedf.reset_index().site.iloc[interp_fac_results==0].to_list()
        print(fsites)
    all_results = lib.interp_from_mf6_depvar_file(depvar_fname,fac_file,"binary",depvar_info["ntime"],"head",interp_thresh,True,
        no_interp_val,usitedf.shape[0])
    datetimes = start_datetime+pd.to_timedelta(all_results["simtime"],unit=model_timeunit)
    allresults_df = pd.DataFrame(all_results["simstate"],index=datetimes,columns=usitedf.index)
    allresults_df.to_csv(depvar_fname+".all.csv")

    if "totim" in obsdf:
        print("WARNING: replacing existing 'totim' column in observation dataframe")
    obsdf.loc[:,"totim"] = obsdf.datetime.apply(lambda x: x  - start_datetime).dt.days 

    usite = obsdf.site.unique()
    usite.sort()
    usite_dict = {s:c for s,c in zip(usite,np.arange(usite.shape[0],dtype=int))}
    obsdf.loc[:,"isite"] = obsdf.site.apply(lambda x: usite_dict[x])
    obsdf.sort_values(by=["isite","totim"],inplace=True)
    
    interp_results = lib.interp_to_obstime(all_results["nproctime"],all_results["simtime"],all_results["simstate"],interp_thresh,"L",
        time_extrap,no_interp_val,obsdf.isite.values,obsdf.totim.values)

    obsdf.loc[:,"simulated"] = interp_results
    lib.uninstall_mf6_grid('grid')
    lib.free_all_memory()
    return {"all_results":allresults_df,"interpolated_results":obsdf}

def get_grid_info_from_gridspec(gridspec_fname: str) -> dict:
    """Read structured grid info from a PEST-style grid specificatin file
    Parameters
    ----------
    gridspec_fname : str
        PEST-style grid specification file
    
    Returns
    -------
    grid_info: dict
        grid information
    """

    if not os.path.exists(gridspec_fname):
        raise FileNotFoundError(gridspec_fname)
    sr = SpatialReference.from_gridspec(gridspec_fname)
    return {
        "x": sr.xcentergrid.flatten(),
        "y": sr.ycentergrid.flatten(),
        "area": sr.areagrid.flatten(),
        "nrow": sr.nrow,
        "ncol": sr.ncol,
        "delr": sr.delr,
        "delc": sr.delc
    }


def get_grid_info_from_mf6_grb(grb_fname: str) -> dict:
    """Read grid info from a MODFLOW-6 binary grid file
    Parameters
    ----------
    grb_fname: str
        MODFLOW-6 binary grid file
    
    Returns
    -------
    grid_info: dict
        grid information
    """
    if not os.path.exists(grb_fname):
        raise FileNotFoundError(grb_fname)
    lib = PestUtilsLib()
    data = lib.install_mf6_grid_from_file("grid",grb_fname)
    data["x"],data["y"],data["z"] = lib.get_cell_centres_mf6("grid",data["ncells"])
    lib.uninstall_mf6_grid("grid")
    lib.free_all_memory()
    return data


def get_2d_grid_info_from_file(fname: str, layer=None) -> dict:
    """Try to read 2-D grid info from a variety of filename sources
    Parameters
    ----------
    fname: str
        filename that stores 2-D grid info.  Optionally, a pandas DataFrame
        at least columns 'x','y' and possibly 'layer'.
    layer: int (optional)
        the (one-based) layer number to use for 2-D.  If None and
        grid info is 3-D, a value of 1 is used
    
    Returns
    -------
    grid_info: dict
        grid information
    """ 

    grid_info = None
    if isinstance(fname,str):
        if not os.path.exists(fname):
            raise FileNotFoundError(fname)
        if fname.lower().endswith(".csv"):
            grid_info = pd.read_csv(fname)
            grid_info.columns = [c.lower() for c in grid_info.columns]
            fname = grid_info # for  checks and processing below
            
        else:
            try:
                grid_info = get_grid_info_from_gridspec(fname)
            except Exception as e1:
                try:
                    grid_info = get_2d_grid_info_from_mf6_grb(fname, layer=layer)
                except Exception as e2:
                    
                    raise Exception("error getting grid info from file '{0}'".format(fname))
        
    if isinstance(fname,pd.DataFrame):
        if 'x' not in fname.columns:
            raise Exception("required 'x' column not found in grid info dataframe")
        if 'y' not in fname.columns:
            raise Exception("required 'y' column not found in grid info dataframe")
        if layer is not None and 'layer' not in fname.columns:
            print("WARNING: 'layer' arg is not None but 'layer' not found in grid info dataframe...")
        # I think these should just be references to column values (not copies)
        grid_info = {c:fname[c].values for c in fname.columns}
    
    return grid_info


def get_2d_grid_info_from_mf6_grb(grb_fname: str,layer=None) -> dict:
    """Read grid info from a MODFLOW-6 binary grid file
    Parameters
    ----------
    grb_fname: str
        MODFLOW-6 binary grid file
    layer: int (optional)
        the layer number to use for 2-D.  If None,
        a value of 1 is used
    
    Returns
    -------
    grid_info: dict
        grid information
    """
    grid_info = get_grid_info_from_mf6_grb(grb_fname)
    nnodes = grid_info["ncells"]
    x = grid_info["x"].copy()
    y = grid_info["y"].copy()
    nrow,ncol = None,None
    if grid_info["idis"] == 1:
        nlay = grid_info["ndim3"]
        if layer is not None:
            assert layer != 0, "Value provided for layer should be 1 based, sorry"
            if layer > nlay:
                raise Exception("user-supplied 'layer' {0} greater than nlay {1}".format(layer,nlay))
        else:
            layer = 1
        nrow = grid_info["ndim2"]
        ncol = grid_info["ndim1"]
        x = x.reshape((nlay,nrow,ncol))[layer-1]
        y = y.reshape((nlay,nrow,ncol))[layer-1]
        grid_info["nnodes"] = nrow * ncol
        grid_info["x"] = x
        grid_info["y"] = y
        grid_info["nrow"] = nrow
        grid_info["ncol"] = ncol

    elif grid_info["idis"] == 2:
        nlay = grid_info["ndim3"]
        if layer is not None:
            if layer > nlay:
                raise Exception("user-supplied 'layer' {0} greater than nlay {1}".format(layer,nlay))
        else:
            layer = 1
        ncpl = grid_info["ndim1"]
        x = x.reshape((nlay,ncpl))[layer-1]
        y = y.reshape((nlay,ncpl))[layer-1]
        grid_info["nnodes"] = ncpl
        grid_info["x"] = x
        grid_info["y"] = y
    return grid_info


def get_2d_pp_info_structured_grid(
    pp_space: int,
    gridinfo_fname: str,
    array_dict = {},
    name_prefix="pp"
) -> pd.DataFrame:
    """Create a grid of pilot point locations for a 
    2-D structured grid
    Parameters
    ----------
    pp_space: int
        row and column spacing for pilot point locations
    gridinfo_fname: str
        file contain grid information
    array_dict: dict (optional)
        a dict of 2-D grid-shape arrays used to populate 
        pilot point attributes.  Special values include:
        "value","zone","bearing","aniso" and "corrlen", 
        although any number of arrays can be passed and will
        sampled at pilot point locations
    name_prefix: str
        pilot point name prefix. Default is "pp"
    
    Returns
    -------
    ppdf: pd.DataaFrame
        dataframe of pilot point information

    """

    grid_info = get_2d_grid_info_from_file(gridinfo_fname)
    pname, px, py, pval = [], [], [], []
    pi, pj = [], []
    parr_dict = {k:[] for k in array_dict.keys()}
    count = 0
    nrow = grid_info["nrow"]
    ncol = grid_info["ncol"]
    nlay = grid_info.get("nlay",1)

    zone_array = array_dict.get("zone",None)

    x = grid_info['x']
    y = grid_info['y']
    x = x.reshape((nlay,nrow,ncol))[0,:,:]
    y = y.reshape((nlay,nrow,ncol))[0,:,:]
    if nrow is None:
        raise Exception("unstructured grid loaded from gridinfo_fname '{0}'".format(gridspec_fname))
    for i in range(int(pp_space / 2), nrow, pp_space):
        for j in range(int(pp_space / 2), ncol, pp_space):
            if zone_array is not None and zone_array[i, j] <= 0:
                continue
            px.append(x[i, j])
            py.append(y[i, j])
            #if zone_array is not None:
            #    pzone.append(zone_array[i, j])
            #else:
            #    pzone.append(1)

            pname.append(name_prefix + "{0}".format(count))
            pi.append(i)
            pj.append(j)
            count += 1
    df = pd.DataFrame(
        {
            "ppname": pname,
            "x": px,
            "y": py,
            "i": pi,
            "j": pj,
        },
        index=pname,
    )
    df.loc[:,"value"] = 1.0
    df.loc[:, "bearing"] = 0.0
    df.loc[:, "aniso"] = 1.0
    delx = pp_space * 5 * int((x.max() - x.min()) / float(ncol))
    dely = pp_space * 5 * int((y.max() - y.min()) / float(nrow))
    df.loc[:, "corrlen"] = max(delx,dely)  # ?
    df.loc[:,"zone"] = 1
    for k,arr in array_dict.items():
        df.loc[:,k] = arr[df.i,df.j]
    df["zone"] = df.zone.astype(int)

    return df


def interpolate_with_sva_pilotpoints_2d(
    pp_info: pandas.DataFrame,
    gridinfo_fname: str,
    vartype="exp",
    krigtype="ordinary",
    vartransform="none",
    max_pts=50,
    min_pts=1,
    search_dist=1e30,
    zone_array=1,
    verbose=True,
    layer=None
) -> dict:
    """Perform 2-D pilot point interpolation using
    spatially varying geostatistical hyper-parameters
    Parameters
    ----------
    pp_info: pandas.DataFrame
        dataframe with pilot point info.  Required columns 
        include: "x","y",and "value".  optional columns include:
        "zone","bearing","aniso",and "corrlen"    
    gridinfo_fname: str
        file name storing grid information
    vartype: str
        variogram type.  Default is "exp"onential
    krigtype: str
        kriging type.  Default is "ordinary"
    vartransform: str
        variogram transformation.  Default is "none"
    max_pts: int
        maximum number of pilot points to use in interpolation.
        Default is 50
    min_pts: int
        minimum number of pilot points to use in interplation.
        Default is 1
    search_dict: float
        search distance to use when looking for nearby pilot points.
        Default is 1.0e+30
    zone_array: int | numpy.ndarray
        the zone array to match up with "zone" value in `pp_info`.  If 
        integer type, a constant zone array of value "zone_array" is used.
        Default is 1
    verbose: bool
        flag to output.  Default is True
    layer: int
        the (one-based) layer number to use if gridinfo_fname points
        to 3-D grid info. Default is None, which results in
        layer 1 being used

    Returns
    -------
    results: dict
        resulting arrays of the various interpolation from pilot 
        points to grid-shaped arrays
    """
    # some checks on pp_info
    req_cols = ["ppname", "x", "y", "value", "corrlen"]
    missing = []
    for req_col in req_cols:
        if req_col not in pp_info.columns:
            missing.append(req_col)
    if len(missing) > 0:
        raise Exception(
            "the following required columns are not in pp_info:{0}".format(
                ",".join(missing)
            )
        )

    if "zone" not in pp_info:
        pp_info.loc[:, "zone"] = 1

    nnodes, nrow, ncol = None, None, None
    x, y, area = None, None, None

    nrow, ncol = None, None
    x, y, area = None, None, None
    grid_info = get_2d_grid_info_from_file(gridinfo_fname, layer)
    nrow = grid_info.get("nrow",None)
    ncol = grid_info.get("ncol",None)
    x = grid_info['x']
    y = grid_info['y']
    area = grid_info.get("area",None)
    idis = grid_info.get("idis",None)
    nnodes = grid_info.get("nnodes",None)
    if area is None:
        area = np.ones_like(x)
    if nnodes is None:
        nnodes = x.shape[0]

    lib = PestUtilsLib()

    if not isinstance(zone_array, np.ndarray):
        zone_array = np.ones((nnodes), dtype=int)
        lib.logger.info("using 1s as zone array for interpolation")
    elif zone_array.dtype != int:
        # TODO warn here
        lib.logger.info("casting zone_array from %r to int", zone_array.dtype)
        zone_array = zone_array.astype(int)
    
    
    hyperfac_ftype = "binary"
    if verbose:
        hyperfac_ftype = "text"
    hyperbearing = 0.0
    hyperaniso = 1.0
    hypervartype = "exp"
    hyperkrigtype = "ordinary"
    hypertrans = "none"

    fac_files = []
    lib.logger.info("using bearing of %r and aniso of %r for hyperpar interpolation", hyperbearing, hyperaniso)
    lib.logger.info("using %r variogram with %r transform for hyperpar interpolation",hypervartype,hypertrans)
       
    results = {}    

    bearing = np.zeros_like(x)
    if "bearing" in pp_info.columns:
        hypernoint = pp_info.bearing.mean()
        lib.logger.info("using no-interpolation value of %r for 'bearing' hyperpar interpolation", hypernoint)
        hyperfac_fname = "tempbearing.fac"
        npts = lib.calc_kriging_factors_auto_2d(
            pp_info.x.values,
            pp_info.y.values,
            pp_info.zone.values.astype(int),
            x.flatten(),
            y.flatten(),
            zone_array.flatten().astype(int),
            hyperkrigtype,
            hyperaniso,
            hyperbearing,
            hyperfac_fname,
            hyperfac_ftype,
        )
        result = lib.krige_using_file(
            hyperfac_fname,
            hyperfac_ftype,
            nnodes,
            hyperkrigtype,
            hypertrans,
            pp_info.bearing.values,
            hypernoint,
            hypernoint,
        )
        bearing = result["targval"]
        fac_files.append(hyperfac_fname)
        if verbose:
            if nrow is not None:
                np.savetxt("bearing.txt",bearing.reshape(nrow,ncol),fmt="%15.6E")
            else:
                np.savetxt("bearing.txt",bearing,fmt="%15.6E")
    results["bearing"] = bearing
    

    aniso = np.ones_like(x)
    if "aniso" in pp_info.columns:
        hypernoint = pp_info.aniso.mean()
        lib.logger.info("using no-interpolation value of %r for 'aniso' hyperpar interpolation", hypernoint)
        
        hyperfac_fname = "tempaniso.fac"
        npts = lib.calc_kriging_factors_auto_2d(
            pp_info.x.values,
            pp_info.y.values,
            pp_info.zone.values,
            x.flatten(),
            y.flatten(),
            zone_array.flatten().astype(int),
            hyperkrigtype,
            hyperaniso,
            hyperbearing,
            hyperfac_fname,
            hyperfac_ftype,
        )
        result = lib.krige_using_file(
            hyperfac_fname,
            hyperfac_ftype,
            nnodes,
            hyperkrigtype,
            hypertrans,
            pp_info.aniso.values,
            hypernoint,
            hypernoint,
        )
        aniso = result["targval"]
        if verbose:
            if nrow is not None:
                np.savetxt("aniso.txt",aniso.reshape(nrow,ncol),fmt="%15.6E")
            else:
                np.savetxt("aniso.txt",aniso,fmt="%15.6E")

    results["aniso"] = aniso

    use_auto = False
    corrlen = None  # todo corrlen default where not in ppdf
    if "corrlen" in pp_info.columns:
        hypernoint = pp_info.corrlen.mean()
        lib.logger.info("using no-interpolation value of %r for 'corrlen' hyperpar interpolation", hypernoint)
        hyperfac_fname = "tempcorrlen.fac"
        npts = lib.calc_kriging_factors_auto_2d(
            pp_info.x.values,
            pp_info.y.values,
            pp_info.zone.values,
            x.flatten(),
            y.flatten(),
            zone_array.flatten().astype(int),
            hyperkrigtype,
            hyperaniso,
            hyperbearing,
            hyperfac_fname,
            hyperfac_ftype,
        )
        result = lib.krige_using_file(
            hyperfac_fname,
            hyperfac_ftype,
            nnodes,
            hyperkrigtype,
            hypertrans,
            pp_info.corrlen.values,
            hypernoint,
            hypernoint,
        )
        corrlen = result["targval"]
        fac_files.append(hyperfac_fname)
        use_auto = False
        if verbose:
            if nrow is not None:
                np.savetxt("corrlen.txt",corrlen.reshape(nrow,ncol),fmt="%15.6E")
            else:
                np.savetxt("corrlen.txt",corrlen,fmt="%15.6E")
        results["corrlen"] = corrlen

    if not verbose:
        for fac_file in fac_files:
            try:
                os.remove(fac_file)
            except Exception as e:
                pass

    # todo: maybe make these args?
    fac_fname = "var.fac"
    fac_ftype = "binary"
    if verbose:
        fac_ftype = "text"
    noint = pp_info.loc[:, "value"].mean()
    if use_auto:
        npts = lib.calc_kriging_factors_auto_2d(
            pp_info.x.values,
            pp_info.y.values,
            pp_info.zone.values,
            x.flatten(),
            y.flatten(),
            zone_array.flatten().astype(int),
            krigtype,
            aniso.flatten(),
            bearing.flatten(),
            fac_fname,
            fac_ftype,
        )
    else:
        npts = lib.calc_kriging_factors_2d(
            pp_info.x.values,
            pp_info.y.values,
            pp_info.zone.values,
            x.flatten(),
            y.flatten(),
            zone_array.flatten().astype(int),
            vartype,
            krigtype,
            corrlen.flatten(),
            aniso.flatten(),
            bearing.flatten(),
            search_dist,
            max_pts,
            min_pts,
            fac_fname,
            fac_ftype,
        )

    result = lib.krige_using_file(
        fac_fname,
        fac_ftype,
        nnodes,
        krigtype,
        vartransform,
        pp_info.loc[:, "value"].values,
        noint,
        noint,
    )
    results["result"] = result["targval"]

    if nrow is not None:
        for k,v in results.items():
            results[k] = v.reshape(nrow,ncol)

    return results


def generate_2d_grid_realizations(
    gridinfo_fname: str,
    num_reals=100,
    variotype="exp",
    mean=1.0,
    variance=1.0,
    variorange=None,
    variotransform="none",
    zone_array=None,
    varioaniso=1.0,
    variobearing=0.0,
    random_seed=12345,
    layer=None,
) ->np.NDArray[float]:
    """draw 2-D realizations using sequential gaussian 
    simulations and optionally using spatially varying 
    geostatistical hyper parameters.
    Parameters
    ----------
    gridinfo_fname: str
        file containing grid information
    num_real : int
        number of realizations to generate
    variotype: str
        variogram type.  Default is "exp"onential
    mean: float or numpy.ndarray
        field mean.  Either a scalar or array of shape nnodes.  Default is 1.0
    variance: float
        field variance. Either a scalar or array of shape nnodes.  Default is 1.0
    variorange: float
        range of the variogram. Either a scalar or array of shape nnodes.
    variotransform: str
        variogram transform.  Default is "none".
    zone_array: int or numpy.ndarray
        the zone array
    varioaniso: float or numpy.ndarray
        the variogram anisotropy ratio.  Either a scalar or array of shape nnodes.
    variobearing: flaot or numpy.ndarray
        the variogram anisotropy bearing.  Either a scalar or array of shape nnodes
    random_seed: int
        the random seed.  Default is 12345
    layer : int or None
        the (one-based) layer to use of gridinfo_fname contains 3-D info.  Default
        is None, which results in layer 1 being used

    Returns
    -------
    results: numpy.ndarray(float)
        realizations (if `grid_info` indicates a structured grid, realizations
        will be reshaped to NROW X NCOL)
    """

    

    nrow, ncol = None, None
    x, y, area = None, None, None
    grid_info = get_2d_grid_info_from_file(gridinfo_fname, layer)
    nrow = grid_info.get("nrow",None)
    ncol = grid_info.get("ncol",None)
    x = grid_info['x']
    y = grid_info['y']
    area = grid_info.get("area",None)
    idis = grid_info.get("idis",None)
    nnodes = grid_info.get("nnodes",None)
    if area is None:
        area = np.ones_like(x)
    if nnodes is None:
        nnodes = x.shape[0]


    if not isinstance(mean, np.ndarray):
        mean = np.zeros((nnodes)) + mean
    if not isinstance(variance, np.ndarray):
        variance = np.zeros((nnodes)) + variance
    if variorange is None:
        delx = x.max() - x.min()
        dely = y.max() - y.min()

        variorange = np.zeros((nnodes)) + max(delx,dely) / 10 # ?
    elif not isinstance(variorange, np.ndarray):
        variorange = np.zeros((nnodes)) + variorange

    if not isinstance(variobearing, np.ndarray):
        variobearing = np.zeros((nnodes)) + variobearing

    if not isinstance(varioaniso, np.ndarray):
        varioaniso = np.zeros((nnodes)) + varioaniso

    if not isinstance(zone_array, np.ndarray):
        zone_array = np.ones((nnodes), dtype=int)
    elif zone_array.dtype != int:
        # TODO warn here
        zone_array = zone_array.astype(int)

    

    power = 1.0

    lib = PestUtilsLib()
    lib.initialize_randgen(random_seed)

    reals = lib.fieldgen2d_sva(
        x.flatten(),
        y.flatten(),
        area.flatten(),
        zone_array.flatten(),
        mean.flatten(),
        variance.flatten(),
        variorange.flatten(),
        varioaniso.flatten(),
        variobearing.flatten(),
        variotransform,
        variotype,
        power,
        num_reals,
    )
    lib.free_all_memory()
    if nrow is not None:
        return reals.transpose().reshape((num_reals, nrow, ncol))
    else:
        return reals.transpose()


class SpatialReference(object):
    """
    a class to locate a structured model grid in x-y space.

    Parameters
    ----------
    delr:  numpy.ndarray
        the model discretization delr vector (An array of spacings along a row)
    delc: numpy ndarray
        the model discretization delc vector (An array of spacings along a column)
    xul: float
        The x coordinate of the upper left corner of the grid. Enter either xul and yul or xll and yll.
    yul: float
        The y coordinate of the upper left corner of the grid. Enter either xul and yul or xll and yll.
    rotation: float
        The counter-clockwise rotation (in degrees) of the grid
    """

    def __init__(self, delr, delc, xul, yul, rotation=0.0):
        self.xul = float(xul)
        self.yul = float(yul)
        self.rotation = float(rotation)
        for delrc in [delr, delc]:
            if isinstance(delrc, float) or isinstance(delrc, int):
                msg = (
                    "delr and delcs must be an array or sequences equal in "
                    "length to the number of rows/columns."
                )
                raise TypeError(msg)

        self.delc = np.atleast_1d(np.array(delc)).astype(np.float64)
        self.delr = np.atleast_1d(np.array(delr)).astype(np.float64)

        self._xgrid = None
        self._ygrid = None
        self._xcentergrid = None
        self._ycentergrid = None

    @property
    def xll(self)->float:
        """lower left x coord
        """
        xll = self.xul - (np.sin(self.theta) * self.yedge[0])
        return xll

    @property
    def yll(self)->float:
        """lower left y coord
        """
        yll = self.yul - (np.cos(self.theta) * self.yedge[0])
        return yll

    @property
    def nrow(self)->int:
        """number of rows
        """
        return self.delc.shape[0]

    @property
    def ncol(self)->int:
        """number of cols
        """
        return self.delr.shape[0]

    @classmethod
    def from_gridspec(cls, gridspec_file)->SpatialReference:
        """instantiate from a pest-style grid specification file
        Parameters
        ----------
        gridspec_file: str
            grid specification file name

        Returns
        -------
        sr: SpatialReference
            sr instance
        """
        f = open(gridspec_file, "r")
        raw = f.readline().strip().split()
        nrow = int(raw[0])
        ncol = int(raw[1])
        raw = f.readline().strip().split()
        xul, yul, rot = float(raw[0]), float(raw[1]), float(raw[2])
        delr = []
        j = 0
        while j < ncol:
            raw = f.readline().strip().split()
            for r in raw:
                if "*" in r:
                    rraw = r.split("*")
                    for n in range(int(rraw[0])):
                        delr.append(float(rraw[1]))
                        j += 1
                else:
                    delr.append(float(r))
                    j += 1
        delc = []
        i = 0
        while i < nrow:
            raw = f.readline().strip().split()
            for r in raw:
                if "*" in r:
                    rraw = r.split("*")
                    for n in range(int(rraw[0])):
                        delc.append(float(rraw[1]))
                        i += 1
                else:
                    delc.append(float(r))
                    i += 1
        f.close()
        return cls(np.array(delr), np.array(delc), xul=xul, yul=yul, rotation=rot)

    @property
    def theta(self)->float:
        """rotation in radians
        """
        return -self.rotation * np.pi / 180.0

    @property
    def xedge(self)->np.NDArray[float]:
        """the xedge array of the grid
        """
        return self.get_xedge_array()

    @property
    def yedge(self)->np.NDArray[float]:
        """the yedge array of the grid
        """
        return self.get_yedge_array()

    @property
    def xgrid(self)->np.NDArray[float]:
        """xgrid array
        """
        if self._xgrid is None:
            self._set_xygrid()
        return self._xgrid

    @property
    def ygrid(self)->np.NDArray[float]:
        """ygrid array
        """
        if self._ygrid is None:
            self._set_xygrid()
        return self._ygrid

    @property
    def xcenter(self)->np.NDArray[float]:
        """grid x center array
        """
        return self.get_xcenter_array()

    @property
    def ycenter(self)->np.NDArray[float]:
        """grid y center array
        """
        return self.get_ycenter_array()

    @property
    def ycentergrid(self)->np.NDArray[float]:
        """grid y center array
        """
        if self._ycentergrid is None:
            self._set_xycentergrid()
        return self._ycentergrid

    @property
    def xcentergrid(self)->np.NDArray[float]:
        """grid x center array
        """
        if self._xcentergrid is None:
            self._set_xycentergrid()
        return self._xcentergrid

    @property
    def areagrid(self)->np.NDArray[float]:
        """area of grid nodes
        """
        dr, dc = np.meshgrid(self.delr, self.delc)
        return dr * dc

    def _set_xycentergrid(self):
        self._xcentergrid, self._ycentergrid = np.meshgrid(self.xcenter, self.ycenter)
        self._xcentergrid, self._ycentergrid = self.transform(
            self._xcentergrid, self._ycentergrid
        )

    def _set_xygrid(self):
        self._xgrid, self._ygrid = np.meshgrid(self.xedge, self.yedge)
        self._xgrid, self._ygrid = self.transform(self._xgrid, self._ygrid)

    def get_xedge_array(self)->np.NDArray[float]:
        """
        a numpy one-dimensional float array that has the cell edge x
        coordinates for every column in the grid in model space - not offset
        or rotated.  Array is of size (ncol + 1)

        """
        assert self.delr is not None and len(self.delr) > 0, (
            "delr not passed to " "spatial reference object"
        )
        xedge = np.concatenate(([0.0], np.add.accumulate(self.delr)))
        return xedge

    def get_yedge_array(self)->np.NDArray[float]:
        """
        a numpy one-dimensional float array that has the cell edge y
        coordinates for every row in the grid in model space - not offset or
        rotated. Array is of size (nrow + 1)

        """
        assert self.delc is not None and len(self.delc) > 0, (
            "delc not passed to " "spatial reference object"
        )
        length_y = np.add.reduce(self.delc)
        yedge = np.concatenate(([length_y], length_y - np.add.accumulate(self.delc)))
        return yedge

    def get_xcenter_array(self)->np.NDArray[float]:
        """
        a numpy one-dimensional float array that has the cell center x
        coordinate for every column in the grid in model space - not offset or rotated.

        """
        assert self.delr is not None and len(self.delr) > 0, (
            "delr not passed to " "spatial reference object"
        )
        x = np.add.accumulate(self.delr) - 0.5 * self.delr
        return x

    def get_ycenter_array(self)->np.NDArray[float]:
        """
        a numpy one-dimensional float array that has the cell center x
        coordinate for every row in the grid in model space - not offset of rotated.

        """
        assert self.delc is not None and len(self.delc) > 0, (
            "delc not passed to " "spatial reference object"
        )
        Ly = np.add.reduce(self.delc)
        y = Ly - (np.add.accumulate(self.delc) - 0.5 * self.delc)
        return y

    @staticmethod
    def rotate(x, y, theta, xorigin=0.0, yorigin=0.0):
        """
        Given x and y array-like values calculate the rotation about an
        arbitrary origin and then return the rotated coordinates.  theta is in
        degrees.

        """
        # jwhite changed on Oct 11 2016 - rotation is now positive CCW
        # theta = -theta * np.pi / 180.
        theta = theta * np.pi / 180.0

        xrot = xorigin + np.cos(theta) * (x - xorigin) - np.sin(theta) * (y - yorigin)
        yrot = yorigin + np.sin(theta) * (x - xorigin) + np.cos(theta) * (y - yorigin)
        return xrot, yrot

    def transform(self, x, y, inverse=False):
        """
        Given x and y array-like values, apply rotation, scale and offset,
        to convert them from model coordinates to real-world coordinates.
        """
        if isinstance(x, list):
            x = np.array(x)
            y = np.array(y)
        if not np.isscalar(x):
            x, y = x.copy(), y.copy()

        if not inverse:
            x += self.xll
            y += self.yll
            x, y = SpatialReference.rotate(
                x, y, theta=self.rotation, xorigin=self.xll, yorigin=self.yll
            )
        else:
            x, y = SpatialReference.rotate(x, y, -self.rotation, self.xll, self.yll)
            x -= self.xll
            y -= self.yll
        return x, y

    def get_extent(self)->tuple[float]:
        """
        Get the extent of the rotated and offset grid

        """
        x0 = self.xedge[0]
        x1 = self.xedge[-1]
        y0 = self.yedge[0]
        y1 = self.yedge[-1]

        # upper left point
        x0r, y0r = self.transform(x0, y0)

        # upper right point
        x1r, y1r = self.transform(x1, y0)

        # lower right point
        x2r, y2r = self.transform(x1, y1)

        # lower left point
        x3r, y3r = self.transform(x0, y1)

        xmin = min(x0r, x1r, x2r, x3r)
        xmax = max(x0r, x1r, x2r, x3r)
        ymin = min(y0r, y1r, y2r, y3r)
        ymax = max(y0r, y1r, y2r, y3r)

        return (xmin, xmax, ymin, ymax)

    def get_vertices(self, i, j)->list[list[float]]:
        """Get vertices for a single cell or sequence if i, j locations."""
        pts = []
        xgrid, ygrid = self.xgrid, self.ygrid
        pts.append([xgrid[i, j], ygrid[i, j]])
        pts.append([xgrid[i + 1, j], ygrid[i + 1, j]])
        pts.append([xgrid[i + 1, j + 1], ygrid[i + 1, j + 1]])
        pts.append([xgrid[i, j + 1], ygrid[i, j + 1]])
        pts.append([xgrid[i, j], ygrid[i, j]])
        if np.isscalar(i):
            return pts
        else:
            vrts = np.array(pts).transpose([2, 0, 1])
            return [v.tolist() for v in vrts]

    def get_ij(self, x, y)->tuple(int):
        """Return the row and column of a point or sequence of points
        in real-world coordinates.

        """
        if np.isscalar(x):
            c = (np.abs(self.xcentergrid[0] - x)).argmin()
            r = (np.abs(self.ycentergrid[:, 0] - y)).argmin()
        else:
            xcp = np.array([self.xcentergrid[0]] * (len(x)))
            ycp = np.array([self.ycentergrid[:, 0]] * (len(x)))
            c = (np.abs(xcp.transpose() - x)).argmin(axis=0)
            r = (np.abs(ycp.transpose() - y)).argmin(axis=0)
        return r, c

    def write_gridspec(self, filename):

        """write a PEST-style grid specification file
        Parameters
        ----------
        filename: str
            file to write


        """
        f = open(filename, "w")
        f.write("{0:10d} {1:10d}\n".format(self.delc.shape[0], self.delr.shape[0]))
        f.write(
            "{0:15.6E} {1:15.6E} {2:15.6E}\n".format(
                self.xul,
                self.yul,
                self.rotation,
            )
        )

        for r in self.delr:
            f.write("{0:15.6E} ".format(r))
        f.write("\n")
        for c in self.delc:
            f.write("{0:15.6E} ".format(c))
        f.write("\n")
        return

